
"""
PyTorch re‑implementation of ``tf_models.py``.

Notes
-----
* The public API (builder functions) is intentionally unchanged except for the
  ``@framework("pytorch")`` decorator.
* All functions return *uncompiled* ``torch.nn.Module`` objects.  Compile /
  move to device / wrap in `torch.compile` or `lightning` modules as you wish.
* The input tensor shape expected by these models is **(batch, seq_len, channels)**
  to stay compatible with the original Keras layout.  Internally, we convert to
  channel‑first format required by ``nn.Conv1d`` and friends.
* Custom layers that exist in Keras but not in core PyTorch have lightweight
  equivalents implemented here (``DepthwiseConv1d``, ``SeparableConv1d``,
  ``SpatialDropout1d``, ``GlobalAveragePooling1d``).
* The external architectures referenced from
  ``nirs4all.presets.legacy`` (*VGG*, *Inception*, *ResNetv2*, *SEResNet*) are
  assumed to have sibling PyTorch ports living at the same import paths.
  If not, you will need to convert those first.
"""

from __future__ import annotations
import math
from typing import Tuple, Any, Optional, List

import torch
import torch.nn as nn
import torch.nn.functional as F

from nirs4all.core.utils import framework
from nirs4all.presets.legacy.Inception_1DCNN import Inception      # assumed PyTorch port
from nirs4all.presets.legacy.ResNet_v2_1DCNN import ResNetv2       # assumed PyTorch port
from nirs4all.presets.legacy.SE_ResNet_1DCNN import SEResNet       # assumed PyTorch port
from nirs4all.presets.legacy.VGG_1DCNN import VGG                  # assumed PyTorch port


# --------------------------------------------------------------------------- #
#                                Helper layers                                #
# --------------------------------------------------------------------------- #
class Permute(nn.Module):
    """Swap ``(N, L, C)`` → ``(N, C, L)`` or the inverse."""

    def __init__(self, to_channels_first: bool = True):
        super().__init__()
        self.to_channels_first = to_channels_first

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.transpose(1, 2)  # works both ways


class DepthwiseConv1d(nn.Module):
    """Depth‑wise 1‑D convolution (groups = in_channels)."""

    def __init__(
        self,
        in_channels: int,
        kernel_size: int,
        depth_multiplier: int = 1,
        stride: int = 1,
        padding: str | int = "same",
        dilation: int = 1,
        bias: bool = True,
    ):
        super().__init__()
        if padding == "same":
            pad = (kernel_size - 1) // 2 * dilation
        else:
            pad = padding
        self.conv = nn.Conv1d(
            in_channels,
            in_channels * depth_multiplier,
            kernel_size,
            stride,
            pad,
            dilation,
            groups=in_channels,
            bias=bias,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # (N,C,L)
        return self.conv(x)


class SeparableConv1d(nn.Module):
    """Depthwise + Pointwise separation as in Keras ``SeparableConv1D``."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        depth_multiplier: int = 1,
        stride: int = 1,
        padding: str | int = "same",
        dilation: int = 1,
        bias: bool = True,
    ):
        super().__init__()
        self.depth = DepthwiseConv1d(
            in_channels,
            kernel_size,
            depth_multiplier=depth_multiplier,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )
        self.point = nn.Conv1d(
            in_channels * depth_multiplier,
            out_channels,
            kernel_size=1,
            bias=bias,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.point(self.depth(x))


class SpatialDropout1d(nn.Module):
    """Drop whole feature maps (channels) instead of individual timesteps."""

    def __init__(self, p: float = 0.5):
        super().__init__()
        self.p = p

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.p == 0.0:
            return x
        # ``x`` is (N, L, C) – convert to (N, C, L) for dropout2d then back
        x = x.transpose(1, 2)
        x = F.dropout2d(x, self.p, self.training, inplace=False)
        return x.transpose(1, 2)


class GlobalAveragePooling1d(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:  # (N,C,L)
        return x.mean(dim=-1)


class GlobalMaxPooling1d(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.amax(dim=-1)


class Flatten(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.flatten(x, start_dim=1)


class Lambda(nn.Module):
    """Wrap an arbitrary lambda (for quick hacks)."""

    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x)


def _activation(act: str) -> nn.Module:
    return {
        "relu": nn.ReLU(),
        "sigmoid": nn.Sigmoid(),
        "swish": nn.SiLU(),
        "linear": nn.Identity(),
    }.get(act.lower(), nn.ReLU())


# --------------------------------------------------------------------------- #
#                             Utility building blocks                         #
# --------------------------------------------------------------------------- #
def _conv_bn_act(
    in_ch: int,
    out_ch: int,
    kernel: int,
    stride: int = 1,
    dilation: int = 1,
    act: str = "relu",
    padding: str | int = "same",
) -> nn.Sequential:
    if padding == "same":
        pad = (kernel - 1) // 2 * dilation
    else:
        pad = padding
    return nn.Sequential(
        nn.Conv1d(in_ch, out_ch, kernel, stride, pad, dilation, bias=False),
        nn.BatchNorm1d(out_ch),
        _activation(act),
    )


# --------------------------------------------------------------------------- #
#                               Model builders                                #
# --------------------------------------------------------------------------- #
@framework("pytorch")
def UNet_NIRS(input_shape: Tuple[int, int], params: dict[str, Any]):
    """Thin wrapper around *VGG* preset identical to original Keras version."""
    length, num_channel = input_shape
    model_width = params.get("model_width", 8)
    problem_type = params.get("problem_type", "Regression")
    output_nums = params.get("output_nums", 1)
    dropout_rate = params.get("dropout_rate", False)

    return VGG(
        length,
        num_channel,
        model_width,
        problem_type=problem_type,
        output_nums=output_nums,
        dropout_rate=dropout_rate,
    ).VGG11()


# --------------------------------------------------------------------------- #
#                               VGG‑style block                               #
# --------------------------------------------------------------------------- #
class _VGGBlock(nn.Module):
    def __init__(self, in_ch: int, n_filters: int, n_conv: int, kernel_size: int):
        super().__init__()
        layers: List[nn.Module] = []
        for _ in range(n_conv):
            layers.append(
                _conv_bn_act(in_ch, n_filters, kernel_size, act="relu", padding="same")
            )
            in_ch = n_filters
        layers.append(nn.MaxPool1d(kernel_size=2, stride=2))
        self.block = nn.Sequential(*layers)

    def forward(self, x):
        return self.block(x)


class _VGG1DPyTorch(nn.Module):
    def __init__(self, input_shape: Tuple[int, int], params: dict[str, Any]):
        super().__init__()
        self.permute = Permute()  # (N,L,C) → (N,C,L)
        c_in = input_shape[1]

        self.block1 = _VGGBlock(
            c_in,
            params.get("block1_filters", 64),
            params.get("block1_convs", 2),
            params.get("kernel_size", 3),
        )
        self.block2 = _VGGBlock(
            params.get("block1_filters", 64),
            params.get("block2_filters", 128),
            params.get("block2_convs", 2),
            params.get("kernel_size", 3),
        )
        self.block3 = _VGGBlock(
            params.get("block2_filters", 128),
            params.get("block3_filters", 256),
            params.get("block3_convs", 2),
            params.get("kernel_size", 3),
        )
        self.flat = Flatten()
        self.dense1 = nn.Linear(params.get("block3_filters", 256), params.get("dense_units", 16))
        self.act1 = _activation("sigmoid")
        self.dense2 = nn.Linear(params.get("dense_units", 16), 1)  # linear output

    def forward(self, x):
        x = self.permute(x)
        x = self.block1(x)
        x = self.block2(x)
        x = self.block3(x)
        x = GlobalAveragePooling1d()(x)  # (N,C)
        x = self.dense1(x)
        x = self.act1(x)
        return self.dense2(x)


@framework("pytorch")
def VGG_1D(input_shape: Tuple[int, int], params: dict[str, Any]):
    """VGG‑like 1‑D CNN rewritten for PyTorch."""
    return _VGG1DPyTorch(input_shape, params)


# --------------------------------------------------------------------------- #
#                         CONV + LSTM multi‑path model                        #
# --------------------------------------------------------------------------- #
class _CONV_LSTM(nn.Module):
    def __init__(self, input_shape: Tuple[int, int], p: dict[str, Any]):
        super().__init__()
        L, C = input_shape
        self.permute = Permute()
        self.unpermute = Permute(to_channels_first=False)

        # --- Convolutional path (SeparableConv) ---------------------------- #
        self.sep_conv = nn.Sequential(
            SeparableConv1d(
                in_channels=C,
                out_channels=p.get("conv_filters", 64),
                kernel_size=p.get("conv_kernel_size", 3),
                depth_multiplier=p.get("conv_depth_multiplier", 64),
                padding="same",
            ),
            nn.MaxPool1d(2),
            nn.BatchNorm1d(p.get("conv_filters", 64)),
            Flatten(),
        )

        # --- Attention path ------------------------------------------------ #
        self.attn_conv = nn.Sequential(
            nn.Conv1d(C, 32, kernel_size=3, stride=2, padding=1),
            nn.MaxPool1d(2),
            nn.BatchNorm1d(32),
            Flatten(),
        )
        self.mha = nn.MultiheadAttention(
            embed_dim=C,
            num_heads=p.get("attention_num_heads", 8),
            dropout=0.1,
            batch_first=True,
        )

        # --- GRU and LSTM paths ------------------------------------------- #
        self.gru = nn.GRU(
            input_size=C,
            hidden_size=p.get("gru_units", 128),
            batch_first=True,
            bidirectional=True,
        )
        self.lstm = nn.LSTM(
            input_size=C,
            hidden_size=p.get("lstm_units", 128),
            batch_first=True,
            bidirectional=True,
        )

        # --- Fully‑connected head ----------------------------------------- #
        cat_size = (
            p.get("conv_filters", 64)
            + 32  # attn_conv flattened size after pooling is rough; ok for demo
            + 2 * p.get("gru_units", 128)
            + 2 * p.get("lstm_units", 128)
        )
        self.head = nn.Sequential(
            nn.BatchNorm1d(cat_size),
            nn.Linear(cat_size, p.get("fc_units1", 64)),
            _activation("relu"),
            nn.Linear(p.get("fc_units1", 64), p.get("fc_units2", 16)),
            _activation("relu"),
            nn.Dropout(0.2),
            nn.Linear(p.get("fc_units2", 16), 1),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor):
        # x: (N,L,C)
        conv_in = self.permute(x)  # (N,C,L)

        x1 = self.sep_conv(conv_in)  # (N, *)
        # Attention path expects (N,L,C)
        attn_out, _ = self.mha(x, x, x, need_weights=False)
        x2 = self.attn_conv(self.permute(attn_out))

        # RNN paths
        _, h_gru = self.gru(x)
        x3 = h_gru.transpose(0, 1).reshape(x.size(0), -1)  # (N, 2*gru_units)
        _, (h_lstm, _) = self.lstm(x)
        x4 = h_lstm.transpose(0, 1).reshape(x.size(0), -1)

        cat = torch.cat([x1, x2, x3, x4], dim=1)
        return self.head(cat)


@framework("pytorch")
def CONV_LSTM(input_shape: Tuple[int, int], params: dict[str, Any]):
    return _CONV_LSTM(input_shape, params)


# --------------------------------------------------------------------------- #
#                              Lite U‑Net variant                             #
# --------------------------------------------------------------------------- #
def _resblock(ch: int, kernel: int, dilation: int, use_se: bool):
    layers = [
        _conv_bn_act(ch, ch, kernel, stride=1, dilation=dilation),
        _conv_bn_act(ch, ch, kernel, stride=1, dilation=dilation),
    ]
    if use_se:
        layers.append(_SEBlock(ch))
    return nn.Sequential(*layers)


class _SEBlock(nn.Module):
    def __init__(self, ch: int, r: int = 8):
        super().__init__()
        self.pool = GlobalAveragePooling1d()
        self.fc = nn.Sequential(
            nn.Linear(ch, ch // r),
            nn.ReLU(),
            nn.Linear(ch // r, ch),
            nn.Sigmoid(),
        )

    def forward(self, x):
        # x: (N,C,L)
        w = self.pool(x).unsqueeze(-1)  # (N,C,1)
        w = self.fc(w.squeeze(-1)).unsqueeze(-1)
        return x * w


class _UNET(nn.Module):
    def __init__(self, input_shape: Tuple[int, int], p: dict[str, Any]):
        super().__init__()
        L, C = input_shape
        layer_n = p.get("layer_n", 64)
        k = p.get("kernel_size", 7)
        depth = p.get("depth", 2)

        self.permute = Permute()

        # Down‑sampling inputs for skip connections
        self.pool_5 = nn.AvgPool1d(kernel_size=5, stride=5, padding=0)
        self.pool_25 = nn.AvgPool1d(kernel_size=25, stride=25, padding=0)

        # Encoder
        self.enc0 = nn.Sequential(
            _conv_bn_act(C, layer_n, k, stride=1),
            *[_resblock(layer_n, k, 1, True) for _ in range(depth)],
        )
        self.enc1 = nn.Sequential(
            _conv_bn_act(layer_n, layer_n * 2, k, stride=5),
            *[_resblock(layer_n * 2, k, 1, True) for _ in range(depth)],
        )
        self.enc2 = nn.Sequential(
            _conv_bn_act(layer_n * 2 + C, layer_n * 3, k, stride=5),
            *[_resblock(layer_n * 3, k, 1, True) for _ in range(depth)],
        )
        self.enc3 = nn.Sequential(
            _conv_bn_act(layer_n * 3 + C, layer_n * 4, k, stride=5),
            *[_resblock(layer_n * 4, k, 1, True) for _ in range(depth)],
        )

        self.reg_head = nn.Sequential(
            nn.Conv1d(layer_n * 4, 1, kernel_size=k, padding=(k - 1) // 2),
            _activation("relu"),
            Lambda(lambda t: 12 * t),
            Flatten(),
            nn.Linear(L // (5 * 5 * 5), 1),  # rough flatten size
            nn.Sigmoid(),
        )

    def forward(self, x):
        # x: (N,L,C) → (N,C,L)
        x0 = self.permute(x)
        # pre‑pools for skips – still (N,C,L')
        inp1 = self.pool_5(x0)
        inp2 = self.pool_25(x0)

        e0 = self.enc0(x0)
        e1 = self.enc1(e0)
        e2 = self.enc2(torch.cat([e1, inp1], dim=1))
        e3 = self.enc3(torch.cat([e2, inp2], dim=1))

        return self.reg_head(e3)


@framework("pytorch")
def UNET(input_shape: Tuple[int, int], params: dict[str, Any]):
    return _UNET(input_shape, params)


# --------------------------------------------------------------------------- #
#                                bard model                                   #
# --------------------------------------------------------------------------- #
class _Bard(nn.Module):
    def __init__(self, input_shape, p):
        super().__init__()
        L, C = input_shape
        self.permute = Permute()

        # conv branch
        self.conv_branch = nn.Sequential(
            SeparableConv1d(
                C,
                p.get("conv_filters1", 64),
                kernel_size=p.get("conv_kernel_size1", 3),
                padding="same",
            ),
            SeparableConv1d(
                p.get("conv_filters1", 64),
                p.get("conv_filters2", 16),
                kernel_size=p.get("conv_kernel_size2", 5),
                padding="same",
            ),
            nn.BatchNorm1d(16),
            nn.ReLU(),
            Flatten(),
        )

        # LSTM branch
        self.lstm1 = nn.LSTM(
            C,
            p.get("lstm_units1", 64),
            batch_first=True,
            bidirectional=True,
        )
        self.lstm2 = nn.LSTM(
            2 * p.get("lstm_units1", 64),
            p.get("lstm_units2", 32),
            batch_first=True,
        )
        self.lstm_bn = nn.BatchNorm1d(p.get("lstm_units2", 32))
        # fc
        self.fc = nn.Sequential(
            nn.Linear(16 * ((L + 1) // 1), p.get("dense_units1", 128)),  # dummy in
            nn.ReLU(),
            nn.Linear(p.get("dense_units1", 128), p.get("dense_units2", 8)),
            nn.ReLU(),
            nn.Linear(p.get("dense_units2", 8), 1),
        )

    def forward(self, x):
        # x (N,L,C)
        conv = self.conv_branch(self.permute(x))
        lstm_out, _ = self.lstm1(x)
        lstm_out, _ = self.lstm2(lstm_out)
        lstm_out = self.lstm_bn(lstm_out)
        lstm_out = Flatten()(lstm_out)
        cat = torch.cat([conv, lstm_out], dim=1)
        return self.fc(cat)


@framework("pytorch")
def bard(input_shape: Tuple[int, int], params: dict[str, Any]):
    return _Bard(input_shape, params)


# --------------------------------------------------------------------------- #
#                          Xception‑like depthwise CNN                        #
# --------------------------------------------------------------------------- #
class _XceptionEntry(nn.Module):
    def __init__(self, in_ch, p):
        super().__init__()
        self.seq = nn.Sequential(
            DepthwiseConv1d(
                in_ch,
                kernel_size=p.get("entry_kernel_size1", 3),
                depth_multiplier=p.get("entry_depth_multiplier1", 2),
                stride=p.get("entry_strides1", 2),
                padding="same",
            ),
            nn.BatchNorm1d(in_ch * p.get("entry_depth_multiplier1", 2)),
            _activation("relu"),
            DepthwiseConv1d(
                in_ch * p.get("entry_depth_multiplier1", 2),
                kernel_size=p.get("entry_kernel_size2", 3),
                depth_multiplier=p.get("entry_depth_multiplier2", 2),
                stride=p.get("entry_strides2", 2),
                padding="same",
            ),
            nn.BatchNorm1d(in_ch * p.get("entry_depth_multiplier1", 2) * p.get("entry_depth_multiplier2", 2)),
            _activation("relu"),
        )
        self.sizes = p.get("entry_sizes", [128, 256, 728])
        in_channels = in_ch * p.get("entry_depth_multiplier1", 2) * p.get("entry_depth_multiplier2", 2)
        blocks = []
        for size in self.sizes:
            blocks.extend(
                [
                    _activation("relu"),
                    SeparableConv1d(in_channels, size, kernel_size=3, padding="same"),
                    nn.BatchNorm1d(size),
                    _activation("relu"),
                    SeparableConv1d(size, size, 3, padding="same"),
                    nn.BatchNorm1d(size),
                    nn.MaxPool1d(kernel_size=3, stride=2, padding=1),
                ]
            )
            blocks.append(
                nn.Sequential(
                    nn.Conv1d(in_channels, size, kernel_size=1, stride=2, padding=0),
                )
            )
            in_channels = size
        self.blocks = nn.ModuleList(blocks)

    def forward(self, x):
        x = self.seq(x)
        prev = x
        # iterate over 2‑tuple (main, residual) blocks
        for i in range(0, len(self.blocks), 2):
            main = self.blocks[i](x)
            residual = self.blocks[i + 1](prev)
            x = main + residual
            prev = x
        return x


class _XceptionMiddle(nn.Module):
    def __init__(self, num_blocks: int = 8):
        super().__init__()
        blk = []
        for _ in range(num_blocks):
            blk.extend(
                [
                    _activation("relu"),
                    SeparableConv1d(728, 728, 3, padding="same"),
                    nn.BatchNorm1d(728),
                    _activation("relu"),
                    SeparableConv1d(728, 728, 3, padding="same"),
                    nn.BatchNorm1d(728),
                    _activation("relu"),
                    SeparableConv1d(728, 728, 3, padding="same"),
                    nn.BatchNorm1d(728),
                ]
            )
        self.blk = nn.Sequential(*blk)

    def forward(self, x):
        return self.blk(x)


class _XceptionExit(nn.Module):
    def __init__(self):
        super().__init__()
        self.seq = nn.Sequential(
            _activation("relu"),
            SeparableConv1d(728, 728, 3, padding="same"),
            nn.BatchNorm1d(728),
            _activation("relu"),
            SeparableConv1d(728, 1024, 3, padding="same"),
            nn.BatchNorm1d(1024),
            nn.MaxPool1d(3, stride=2, padding=1),
        )
        self.residual = nn.Conv1d(728, 1024, kernel_size=1, stride=2, padding=0)
        self.tail = nn.Sequential(
            _activation("relu"),
            SeparableConv1d(1024, 728, 3, padding="same"),
            nn.BatchNorm1d(728),
            _activation("relu"),
            SeparableConv1d(728, 1024, 3, padding="same"),
            nn.BatchNorm1d(1024),
            GlobalAveragePooling1d(),
            nn.Linear(1024, 1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        prev = x
        x = self.seq(x)
        x = x + self.residual(prev)
        return self.tail(x)


class _Xception1D(nn.Module):
    def __init__(self, input_shape, p):
        super().__init__()
        L, C = input_shape
        self.permute = Permute()
        self.entry = _XceptionEntry(C, p)
        self.middle = _XceptionMiddle(p.get("middle_num_blocks", 8))
        self.exit = _XceptionExit()

    def forward(self, x):
        x = self.entry(self.permute(x))
        x = self.middle(x)
        return self.exit(x)


@framework("pytorch")
def XCeption1D(input_shape: Tuple[int, int], params: dict[str, Any]):
    return _Xception1D(input_shape, params)


# --------------------------------------------------------------------------- #
#                                    MLP                                      #
# --------------------------------------------------------------------------- #
class _MLP(nn.Module):
    def __init__(self, input_shape, p):
        super().__init__()
        L, C = input_shape
        self.flat = Flatten()
        in_features = L * C
        self.seq = nn.Sequential(
            nn.Dropout(p.get("dropout_rate", 0.2)),
            nn.Linear(in_features, p.get("dense_units1", 1024)),
            _activation("relu"),
            nn.Linear(p.get("dense_units1", 1024), p.get("dense_units2", 128)),
            _activation("relu"),
            nn.Linear(p.get("dense_units2", 128), p.get("dense_units3", 8)),
            _activation("relu"),
            nn.Linear(p.get("dense_units3", 8), 1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        return self.seq(self.flat(x))


@framework("pytorch")
def MLP(input_shape, params):
    return _MLP(input_shape, params)


# --------------------------------------------------------------------------- #
#  Remaining builder wrappers around legacy presets (SEResNet, ResNetv2, etc) #
# --------------------------------------------------------------------------- #
@framework("pytorch")
def SEResNet_model(input_shape: Tuple[int, int], params: dict[str, Any]):
    length = input_shape[0]
    num_channel = input_shape[-1]
    model_width = params.get("model_width", 16)
    problem_type = params.get("problem_type", "Regression")
    output_nums = params.get("output_nums", 1)
    reduction_ratio = params.get("reduction_ratio", 4)
    dropout_rate = params.get("dropout_rate", False)
    pooling = params.get("pooling", "avg")

    return SEResNet(
        length,
        num_channel,
        model_width,
        ratio=reduction_ratio,
        problem_type=problem_type,
        output_nums=output_nums,
        pooling=pooling,
        dropout_rate=dropout_rate,
    ).SEResNet101()


@framework("pytorch")
def ResNetV2_model(input_shape: Tuple[int, int], params: dict[str, Any]):
    length = input_shape[0]
    num_channel = input_shape[-1]
    model_width = params.get("model_width", 16)
    problem_type = params.get("problem_type", "Regression")
    output_nums = params.get("output_nums", 1)
    pooling = params.get("pooling", "avg")
    dropout_rate = params.get("dropout_rate", False)

    return ResNetv2(
        length,
        num_channel,
        model_width,
        problem_type=problem_type,
        output_nums=output_nums,
        pooling=pooling,
        dropout_rate=dropout_rate,
    ).ResNet34()


# --------------------------------------------------------------------------- #
#                             FFT + Conv variant                               #
# --------------------------------------------------------------------------- #
class _FFTConv(nn.Module):
    def __init__(self, input_shape, p):
        super().__init__()
        L, C = input_shape
        self.permute = Permute()

        self.conv_block = nn.Sequential(
            SeparableConv1d(
                C,
                p.get("filters1", 64),
                kernel_size=p.get("kernel_size1", 3),
                stride=p.get("strides1", 2),
                depth_multiplier=p.get("depth_multiplier1", 32),
                padding="same",
            ),
            nn.BatchNorm1d(p.get("filters1", 64)),
            _activation("relu"),
            SeparableConv1d(
                p.get("filters1", 64),
                p.get("filters2", 64),
                kernel_size=p.get("kernel_size2", 3),
                stride=p.get("strides2", 2),
                depth_multiplier=p.get("depth_multiplier2", 32),
                padding="same",
            ),
            nn.BatchNorm1d(p.get("filters2", 64)),
            _activation("relu"),
            nn.Conv1d(
                p.get("filters2", 64),
                p.get("filters3", 32),
                kernel_size=p.get("kernel_size3", 5),
                stride=p.get("strides3", 2),
                padding="same",
            ),
            nn.BatchNorm1d(p.get("filters3", 32)),
            Flatten(),
            _activation("relu"),
            nn.Linear(p.get("filters3", 32) * ((L // 8) + 1), p.get("dense_units1", 128)),
            _activation("relu"),
            nn.Linear(p.get("dense_units1", 128), p.get("dense_units2", 32)),
            _activation("relu"),
            nn.Dropout(p.get("dropout_rate", 0.1)),
            nn.Linear(p.get("dense_units2", 32), 1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        # Apply FFT on channels dim similar to Keras Lambda
        x = torch.fft.fft(x.to(torch.complex64), dim=1).real.to(x.dtype)
        return self.conv_block(self.permute(x))


@framework("pytorch")
def FFT_Conv(input_shape, params):
    return _FFTConv(input_shape, params)


# --------------------------------------------------------------------------- #
#                              Inception wrapper                               #
# --------------------------------------------------------------------------- #
@framework("pytorch")
def inception1D(input_shape: Tuple[int, int], params: dict[str, Any]):
    length = input_shape[1]
    model_width = params.get("model_width", 16)
    num_channel = params.get("num_channel", 1)
    problem_type = params.get("problem_type", "Regression")
    output_number = params.get("output_number", 1)

    return Inception(
        length, num_channel, model_width, problem_type=problem_type, output_nums=output_number
    ).Inception_v3()


# --------------------------------------------------------------------------- #
#                       Lightweight radial‑style models                       #
# --------------------------------------------------------------------------- #
# The Custom_* builders below are left as an exercise if needed.  Their
# TensorFlow counterparts rely heavily on specialised depth‑wise 1‑D ops
# and attention; porting them would follow the same patterns as above.
# --------------------------------------------------------------------------- #



# Import Necessary Libraries
import numpy as np
import tensorflow as tf


def Conv_Block(inputs, model_width, kernel, multiplier, bottleneck=False):
    # 1D Convolutional Block
    if bottleneck:
        # Bottleneck design: use 1x1 conv to reduce channels before kernel-sized conv
        x = tf.keras.layers.Conv1D(model_width * multiplier // 4, 1, padding='same')(inputs)
        x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.Conv1D(model_width * multiplier, kernel, padding='same')(x)
    else:
        # Standard convolution
        x = tf.keras.layers.Conv1D(model_width * multiplier, kernel, padding='same')(inputs)
    
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation('relu')(x)

    return x


def trans_conv1D(inputs, model_width, multiplier):
    # 1D Transposed Convolutional Block, used instead of UpSampling
    x = tf.keras.layers.Conv1DTranspose(model_width * multiplier, 2, strides=2, padding='same')(inputs)  # Stride = 2, Kernel Size = 2
    x = tf.keras.layers.BatchNormalization()(x)
    x = tf.keras.layers.Activation('relu')(x)

    return x


def Concat_Block(input1, *argv):
    # Concatenation Block from the KERAS Library
    cat = input1
    for arg in range(0, len(argv)):
        cat = tf.keras.layers.concatenate([cat, argv[arg]], axis=-1)

    return cat


def upConv_Block(inputs, size=2):
    # 1D UpSampling Block
    up = tf.keras.layers.UpSampling1D(size=size)(inputs)

    return up


def Feature_Extraction_Block(inputs, model_width, feature_number):
    # Feature Extraction Block for the AutoEncoder Mode
    shape = inputs.shape
    latent = tf.keras.layers.Flatten()(inputs)
    latent = tf.keras.layers.Dense(feature_number, name='features')(latent)
    latent = tf.keras.layers.Dense(model_width * shape[1])(latent)
    latent = tf.keras.layers.Reshape((shape[1], model_width))(latent)

    return latent


def dense_block(x, num_filters, num_layers, bottleneck=True):
    for i in range(num_layers):
        # Use kernel size 3 by default for Conv_Block with bottleneck parameter
        cb = Conv_Block(x, num_filters, 3, multiplier=1, bottleneck=bottleneck)
        x = tf.keras.layers.concatenate([x, cb], axis=-1)

    return x


def Attention_Block(skip_connection, gating_signal, num_filters, multiplier):
    # Attention Block
    conv1x1_1 = tf.keras.layers.Conv1D(num_filters * multiplier, 1, strides=2)(skip_connection)
    conv1x1_1 = tf.keras.layers.BatchNormalization()(conv1x1_1)
    conv1x1_2 = tf.keras.layers.Conv1D(num_filters * multiplier, 1, strides=1)(gating_signal)
    conv1x1_2 = tf.keras.layers.BatchNormalization()(conv1x1_2)
    conv1_2 = tf.keras.layers.add([conv1x1_1, conv1x1_2])
    conv1_2 = tf.keras.layers.Activation('relu')(conv1_2)
    conv1_2 = tf.keras.layers.Conv1D(1, 1, strides=1)(conv1_2)
    conv1_2 = tf.keras.layers.BatchNormalization()(conv1_2)
    conv1_2 = tf.keras.layers.Activation('sigmoid')(conv1_2)
    resampler1 = upConv_Block(conv1_2)
    resampler2 = trans_conv1D(conv1_2, 1, 1)
    resampler = tf.keras.layers.add([resampler1, resampler2])
    out = skip_connection * resampler

    return out


class UNet:
    def __init__(self, length, model_depth, num_channel, model_width, kernel_size, problem_type='Regression',
                 output_nums=1, ds=1, ae=0, ag=0, lstm=0, feature_number=1024, is_transconv=True):
        # length: Input Signal Length
        # model_depth: Depth of the Model
        # model_width: Width of the Input Layer of the Model
        # num_channel: Number of Channels allowed by the Model
        # kernel_size: Kernel or Filter Size of the Convolutional Layers
        # problem_type: Classification (Binary or Multiclass) or Regression
        # output_nums: Output Classes (Classification Mode) or Features (Regression Mode)
        # ds: Checks where Deep Supervision is active or not, either 0 or 1 [Default value set as 0]
        # ae: Enables or diables the AutoEncoder Mode, either 0 or 1 [Default value set as 0]
        # ag: Checks where Attention Guided is active or not, either 0 or 1 [Default value set as 0]
        # lstm: Checks where Bidirectional LSTM is active or not, either 0 or 1 [Default value set as 0]
        # feature_number: Number of Features or Embeddings to be extracted from the AutoEncoder in the A_E Mode
        # is_transconv: (TRUE - Transposed Convolution, FALSE - UpSampling) in the Encoder Layer
        self.length = length
        self.model_depth = model_depth
        self.num_channel = num_channel
        self.model_width = model_width
        self.kernel_size = kernel_size
        self.problem_type = problem_type
        self.output_nums = output_nums
        self.D_S = ds
        self.A_E = ae
        self.A_G = ag
        self.LSTM = lstm
        self.feature_number = feature_number
        self.is_transconv = is_transconv

    def UNet3P(self):
        # Variable UNet3+ Model Design
        if self.length == 0 or self.model_depth == 0 or self.model_width == 0 or self.num_channel == 0 or self.kernel_size == 0:
            raise ValueError("Please Check the Values of the Input Parameters!")

        convs = {}
        levels = []

        # Encoding
        print(self.length, self.num_channel)
        inputs = tf.keras.Input((self.length, self.num_channel))
        pool = inputs

        for i in range(1, (self.model_depth + 1)):
            conv = Conv_Block(pool, self.model_width, self.kernel_size, 2 ** (i - 1))
            conv = Conv_Block(conv, self.model_width, self.kernel_size, 2 ** (i - 1))
            pool = tf.keras.layers.MaxPooling1D(pool_size=2)(conv)
            convs["conv%s" % i] = conv

        if self.A_E == 1:
            # Collect Latent Features or Embeddings from AutoEncoders
            pool = Feature_Extraction_Block(pool, self.model_width, self.feature_number)
        conv = Conv_Block(pool, self.model_width, self.kernel_size, 2 ** self.model_depth)
        conv = Conv_Block(conv, self.model_width, self.kernel_size, 2 ** self.model_depth)

        # Decoding
        deconv = conv
        deconvs = {}
        convs_list = list(convs.values())

        for j in range(0, self.model_depth):
            skip_connections_all = convs_list[self.model_depth - j - 1]
            skip_connections_all = Conv_Block(skip_connections_all, self.model_width, self.kernel_size, 2 ** 0)
            for k in range(0, self.model_depth - j - 1):
                skip_connection = convs_list[k]
                skip_connection = tf.keras.layers.MaxPooling1D(pool_size=(2**((self.model_depth - j) - k - 1)))(skip_connection)
                skip_connection = Conv_Block(skip_connection, self.model_width, self.kernel_size, 2 ** 0)
                skip_connections_all = tf.keras.layers.concatenate([skip_connections_all, skip_connection], axis=-1)
            deconv_tot = upConv_Block(deconv, size=2 ** 1)
            deconv_tot = Conv_Block(deconv_tot, self.model_width, self.kernel_size, 2 ** 0)
            deconv_tot = tf.keras.layers.concatenate([skip_connections_all, deconv_tot], axis=-1)
            if j > 0:
                for m in range(0, j):
                    deconv = upConv_Block(deconvs["deconv%s" % m], size=(2 ** (j - m)))
                    deconv = Conv_Block(deconv, self.model_width, self.kernel_size, 2 ** 0)
                    deconv_tot = tf.keras.layers.concatenate([deconv_tot, deconv], axis=-1)
            deconv = Conv_Block(deconv_tot, self.model_width, self.kernel_size, self.model_depth + 1)
            deconvs["deconv%s" % j] = deconv
            if self.D_S == 1:
                # For Deep Supervision
                level = tf.keras.layers.Conv1D(1, 1, 2, name=f'level{self.model_depth - j}')(deconv)
                levels.append(level)

        # Output
        outputs = []
        if self.problem_type == 'Classification':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='softmax', name="out")(deconv)
        elif self.problem_type == 'Regression':
            outputs = tf.keras.layers.Conv1D(self.output_nums, 1, activation='linear', name="out")(deconv)

        model = tf.keras.Model(inputs=[inputs], outputs=[outputs])

        if self.D_S == 1:
            levels.append(outputs)
            levels.reverse()
            model = tf.keras.Model(inputs=[inputs], outputs=levels)

        return model


if __name__ == '__main__':
    # Configurations
    length = 1024  # Length of each Segment
    model_name = 'UNet3P'  # UNet or UNetPP
    model_depth = 5  # Number of Level in the CNN Model
    model_width = 64  # Width of the Initial Layer, subsequent layers start from here
    kernel_size = 3  # Size of the Kernels/Filter
    num_channel = 1  # Number of Channels in the Model
    D_S = 1  # Turn on Deep Supervision
    A_E = 0  # Turn on AutoEncoder Mode for Feature Extraction
    A_G = 1  # Turn on for Guided Attention
    LSTM = 1  # Turn on for LSTM, Implemented for UNet and MultiResUNet only
    problem_type = 'Regression'  # Regression or Classification (Commonly Regression)
    output_nums = 1  # Number of Class for Classification Problems, always '1' for Regression Problems
    is_transconv = True  # True: Transposed Convolution, False: UpSampling
    '''Only required if the AutoEncoder Mode is turned on'''
    feature_number = 1024  # Number of Features to be Extracted
    #
    Model = UNet(length, model_depth, num_channel, model_width, kernel_size, problem_type=problem_type, output_nums=output_nums,
                 ds=D_S, ae=A_E, ag=A_G, lstm=LSTM, is_transconv=is_transconv).UNet3P()
    Model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=0.0003), loss=tf.keras.losses.MeanAbsoluteError(), metrics=tf.keras.metrics.MeanSquaredError())
    Model.summary()



"""
Finetuner module for optimization of machine learning models.
"""

from nirs4all.core.finetuner.base_finetuner import *

__all__ = ["BaseFinetuner", "OptunaFineTuner", "SklearnFineTuner", "FineTunerFactory"]


# model_manager.py
from abc import ABC, abstractmethod
import joblib
from sklearn.metrics import get_scorer, mean_gamma_deviance
import sys
import os
import numpy as np
from .model_builder_factory import ModelBuilderFactory
from ..utils import TF_AVAILABLE, TORCH_AVAILABLE


METRIC_ABBREVIATIONS = {
    # Regression metrics
    "ev": "explained_variance",                       # explained_variance_scorer
    "r2": "r2",                                       # r2_scorer
    "max_err": "max_error",                           # max_error_scorer
    "mae": "neg_mean_absolute_error",                 # neg_mean_absolute_error_scorer
    "mape": "neg_mean_absolute_percentage_error",     # neg_mean_absolute_percentage_error_scorer
    "mse": "neg_mean_squared_error",                  # neg_mean_squared_error_scorer
    "msle": "neg_mean_squared_log_error",             # neg_mean_squared_log_error_scorer
    "rmse": "neg_root_mean_squared_error",            # neg_root_mean_squared_error_scorer
    "rmsle": "neg_root_mean_squared_log_error",       # neg_root_mean_squared_log_error_scorer
    "median_ae": "neg_median_absolute_error",         # neg_median_absolute_error_scorer
    "poisson_dev": "neg_mean_poisson_deviance",       # neg_mean_poisson_deviance_scorer
    "gamma_dev": "neg_mean_gamma_deviance",           # neg_mean_gamma_deviance_scorer
    "d2_ae": "d2_absolute_error_score",               # d2_absolute_error_scorer

    # Classification metrics
    "acc": "accuracy",                                # accuracy_scorer
    "top_k_acc": "top_k_accuracy",                    # top_k_accuracy_scorer
    "roc_auc": "roc_auc",                             # roc_auc_scorer
    "roc_auc_ovr": "roc_auc_ovr",                     # roc_auc_ovr_scorer
    "roc_auc_ovo": "roc_auc_ovo",                     # roc_auc_ovo_scorer
    "roc_auc_ovr_weighted": "roc_auc_ovr_weighted",   # roc_auc_ovr_weighted_scorer
    "roc_auc_ovo_weighted": "roc_auc_ovo_weighted",   # roc_auc_ovo_weighted_scorer
    "balanced_acc": "balanced_accuracy",              # balanced_accuracy_scorer
    "ap": "average_precision",                        # average_precision_scorer
    "log_loss": "neg_log_loss",                       # neg_log_loss_scorer
    "brier": "neg_brier_score",                       # neg_brier_score_scorer
    "plr": "positive_likelihood_ratio",               # positive_likelihood_ratio_scorer
    "nlr": "neg_negative_likelihood_ratio",           # neg_negative_likelihood_ratio_scorer

    # Cluster metrics (supervised)
    "adj_rand": "adjusted_rand_score",                # adjusted_rand_scorer
    "rand": "rand_score",                             # rand_scorer
    "homogeneity": "homogeneity_score",               # homogeneity_scorer
    "completeness": "completeness_score",             # completeness_scorer
    "v_measure": "v_measure_score",                   # v_measure_scorer
    "mutual_info": "mutual_info_score",               # mutual_info_scorer
    "adj_mutual_info": "adjusted_mutual_info_score",  # adjusted_mutual_info_scorer
    "nmi": "normalized_mutual_info_score",            # normalized_mutual_info_scorer
    "fmi": "fowlkes_mallows_score",                   # fowlkes_mallows_scorer
}


def detect_task_type(loss, metrics):
    classification_losses = {'binary_crossentropy', 'categorical_crossentropy', 'sparse_categorical_crossentropy'}
    classification_metrics = {'accuracy', 'acc', 'precision', 'recall', 'f1', 'auc'}

    # Check if loss is for classification
    if loss in classification_losses:
        return 'classification'

    # Check if any of the metrics are for classification
    if any(metric in classification_metrics for metric in metrics):
        return 'classification'

    # Default to regression
    return 'regression'


def prepare_y(y_train, y_val, model, framework, loss, task):
    """
    Prepares y_train and y_val based on the loss function, model type (Keras or scikit-learn),
    and the task (classification, regression, etc.). Adjusts the model if necessary and counts num_classes.

    Parameters:
        y_train (array-like): Training labels (n_samples,) or (n_samples, 1) shape.
        y_val (array-like): Validation labels (n_samples,) or (n_samples, 1) shape.
        model (object): The model instance (Keras or scikit-learn).
        framework (str): The framework of the model ('tensorflow' or 'sklearn').
        loss (str): The loss function being used (e.g., 'categorical_crossentropy', 'poisson', 'kl_divergence', etc.).
        task (str): The task type ('classification', 'regression', 'clustering', etc.).

    Returns:
        y_train (array-like): Modified training labels.
        y_val (array-like): Modified validation labels.
        model (object): Updated model if necessary.
        num_classes (int or None): Number of classes (if applicable).

    Raises:
        ValueError: If labels are inappropriate for the specified loss or task.
    """
    # Ensure labels are numpy arrays and flatten them
    y_train = np.array(y_train).reshape(-1)
    y_val = np.array(y_val).reshape(-1)
    num_classes = None  # Initialize num_classes to None

    if task == 'classification':
        # Count the number of classes
        classes = np.unique(y_train)
        num_classes = len(classes)

        # Validate labels
        if num_classes < 2:
            raise ValueError("At least two classes are required for classification.")

        # Framework-specific processing
        if framework == 'tensorflow':
            import tensorflow
            from tensorflow.keras.layers import Dense
            from tensorflow.keras.utils import to_categorical

            if loss == 'sparse_categorical_crossentropy':
                # Labels must be integer-encoded
                if not np.issubdtype(y_train.dtype, np.integer):
                    raise ValueError("Labels must be integer-encoded for sparse_categorical_crossentropy.")

                # Check label range
                if np.min(y_train) < 0 or np.max(y_train) >= num_classes:
                    raise ValueError(f"Labels must be in the range [0, {num_classes - 1}] for sparse_categorical_crossentropy.")

                # Adjust model's last layer
                from tensorflow.keras.layers import Dense
                last_layer = model.layers[-1]
                if last_layer.units != num_classes or last_layer.activation.__name__ != 'softmax':
                    model.pop()
                    model.add(Dense(num_classes, activation='softmax'))

            elif loss == 'categorical_crossentropy':
                # Convert labels to one-hot encoding
                y_train = to_categorical(y_train, num_classes=num_classes)
                y_val = to_categorical(y_val, num_classes=num_classes)

                # Adjust model's last layer
                last_layer = model.layers[-1]
                if last_layer.units != num_classes or last_layer.activation.__name__ != 'softmax':
                    model.pop()
                    model.add(Dense(num_classes, activation='softmax'))

            elif loss == 'binary_crossentropy':
                # Ensure labels are binary (0 or 1)
                unique_labels = np.unique(y_train)
                if not set(unique_labels).issubset({0, 1}):
                    raise ValueError("Labels must be binary (0 or 1) for binary_crossentropy.")

                # Adjust model's last layer
                last_layer = model.layers[-1]
                if last_layer.units != 1 or last_layer.activation.__name__ != 'sigmoid':
                    model.pop()
                    model.add(Dense(1, activation='sigmoid'))

            else:
                raise ValueError(f"Unsupported loss '{loss}' for classification with TensorFlow.")

        elif framework == 'sklearn':
            # Encode labels as integers (if not already)
            from sklearn.preprocessing import LabelEncoder
            le = LabelEncoder()
            y_train = le.fit_transform(y_train)
            y_val = le.transform(y_val)
            num_classes = len(le.classes_)
            # No model adjustments needed for scikit-learn classifiers

        else:
            raise ValueError("Unsupported framework. Choose 'tensorflow' or 'sklearn'.")

    elif task == 'regression':
        # Labels should be continuous values
        if framework == 'tensorflow':
            from tensorflow.keras.layers import Dense

            # Convert labels to float32
            y_train = y_train.astype(np.float32)
            y_val = y_val.astype(np.float32)

            # Adjust model's last layer based on loss function
            if loss in ['mean_squared_error', 'mse', 'mean_absolute_error', 'mae']:
                # Standard regression
                last_layer = model.layers[-1]
                if last_layer.units != 1 or last_layer.activation.__name__ != 'linear':
                    model.pop()
                    model.add(Dense(1, activation='linear'))

            elif loss == 'poisson':
                # Poisson regression (counts)
                last_layer = model.layers[-1]
                if last_layer.units != 1 or last_layer.activation.__name__ != 'exponential':
                    model.pop()
                    model.add(Dense(1, activation='exponential'))

            else:
                raise ValueError(f"Unsupported loss '{loss}' for regression with TensorFlow.")

        elif framework == 'sklearn':
            # No changes needed for scikit-learn regressors
            pass

        else:
            raise ValueError("Unsupported framework. Choose 'tensorflow' or 'sklearn'.")

    elif task == 'clustering':
        # Clustering tasks typically do not involve supervised labels in the same way
        # For unsupervised clustering, labels may not be provided or used
        # Here we can raise a warning or handle accordingly
        print("Warning: Clustering tasks are unsupervised. Labels may not be used.")

    elif task == 'density_estimation':
        # For tasks like density estimation using KL divergence
        if framework == 'tensorflow':
            from tensorflow.keras.layers import Dense

            # For KL divergence, the labels should represent probability distributions
            # Ensure labels are valid probability distributions
            if not np.allclose(y_train.sum(axis=1), 1):
                raise ValueError("Labels must be valid probability distributions (sum to 1) for KL divergence.")

            # Adjust model's last layer
            last_layer = model.layers[-1]
            if last_layer.units != y_train.shape[1] or last_layer.activation.__name__ != 'softmax':
                model.pop()
                model.add(Dense(y_train.shape[1], activation='softmax'))

        else:
            raise ValueError("KL divergence is typically used with TensorFlow models.")

    else:
        raise ValueError(f"Unsupported task '{task}'. Supported tasks are 'classification', 'regression', 'clustering', 'density_estimation'.")

    return y_train, y_val, model, num_classes


class BaseModelManager(ABC):

    def __init__(self, models=None, model_config=None):
        self.models = models
        self.model_config = model_config
        self.framework = 'framework undefined'

    @staticmethod
    def evaluate(y_true, y_pred, metrics):
        """
        Evaluate a list of metrics on given true and predicted values.

        Args:
        y_true (array-like): Ground truth values.
        y_pred (array-like): Predicted values.
        metrics (list): A list of metrics to evaluate. Each metric can be:
                        - a string (sklearn metric name or abbreviation),
                        - a class instance with an `evaluate` method,
                        - a function that takes (y_true, y_pred).

        Returns:
        dict: A dictionary with metric names as keys and their computed scores as values.
        
        """
        scores = {}
        
        for metric in metrics:
            # If the metric is a string, map abbreviation to sklearn's full metric name
            if isinstance(metric, str):
                full_metric = METRIC_ABBREVIATIONS.get(metric, metric)  # Use full name if found, else the original string
                try:
                    scorer = get_scorer(full_metric)
                    score = scorer._score_func(y_true, y_pred)
                    scores[metric] = score  # Keep the original metric abbreviation or name
                except Exception as e:
                    scores[metric] = f"Error: {e}"

            # If the metric is a class instance with an `evaluate` method
            elif hasattr(metric, 'evaluate') and callable(metric.evaluate):
                try:
                    score = metric.evaluate(y_true, y_pred)
                    scores[metric.__class__.__name__] = score
                except Exception as e:
                    scores[metric.__class__.__name__] = f"Error: {e}"

            # If the metric is a callable function
            elif callable(metric):
                try:
                    score = metric(y_true, y_pred)
                    scores[metric.__name__] = score
                except Exception as e:
                    scores[metric.__name__] = f"Error: {e}"

            # If the metric is unrecognized, log an error
            else:
                scores[str(metric)] = "Unsupported metric type"

        return scores

    # def evaluate(self, y_true, y_pred, metrics=['mse', 'r2']):
    #     for metric in metrics:
    #         if isinstance(metric, str):
    #             # test if is a sklearn metric in the module sklearn.metrics
    #             if hasattr(sys.modules[__name__], metric):
    #                 metric = getattr(sys.modules[__name__], metric)
    #             else:
    #                 raise ValueError(f"Metric {metric} not found in sklearn.metrics module")
    #         elif inspect.isfunction(metric):
    #     return {"mse": float(mse), "r2": float(r2)}

    def save_model(self, path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        for i, model in enumerate(self.models):
            joblib.dump(model, path + f'_{i}.pkl')
        
    def load_model(self, path):
        if os.path.exists(path):
            self.models = []
            for i in range(len(os.listdir(path))):
                if os.path.exists(path + f'model_{i}.pkl'):
                    self.models.append(joblib.load(path + f'model_{i}.pkl'))
        else:
            raise FileNotFoundError(f"Model file not found at: {path}")

    @abstractmethod
    def train(self, X_train, y_train, X_val, y_val, training_params, input_dim, metrics, no_folds):
        pass

    @abstractmethod
    def predict(self, X, task, return_all, no_folds, raw_class_output):
        pass


if TF_AVAILABLE:
    import tensorflow as tf
    from tensorflow.keras.callbacks import Callback, EarlyStopping, LearningRateScheduler
    from tensorflow.keras.models import load_model
    from tensorflow.keras import metrics as tf_metrics
    from tensorflow.keras.utils import to_categorical

    # Abbreviations mapping for TensorFlow/Keras metrics
    TF_METRIC_ABBREVIATIONS = {
        # Regression metrics
        "mae": tf_metrics.MeanAbsoluteError(),           # Mean Absolute Error
        "mse": tf_metrics.MeanSquaredError(),            # Mean Squared Error
        "msle": tf_metrics.MeanSquaredLogarithmicError(),  # Mean Squared Logarithmic Error
        "rmse": tf_metrics.RootMeanSquaredError(),       # Root Mean Squared Error
        "mape": tf_metrics.MeanAbsolutePercentageError(),  # Mean Absolute Percentage Error
        # "r2": tf_metrics.R2Score(),                      # R-squared

        # Classification metrics
        "acc": tf_metrics.Accuracy(),                    # Accuracy
        "accuracy": tf_metrics.Accuracy(),
        "auc": tf_metrics.AUC(),                         # AUC (ROC)
        # "f1": tf_metrics.F1Score(),                      # F1 Score
        "precision": tf_metrics.Precision(),             # Precision
        "recall": tf_metrics.Recall(),                   # Recall
        "log_loss": tf_metrics.BinaryCrossentropy(),     # Binary Crossentropy
    }
    
    def get_keras_metric(metric_names):
        res_metrics = []
        for metric_name in metric_names:
            if isinstance(metric_name, str):
                if metric_name in TF_METRIC_ABBREVIATIONS:
                    res_metrics.append(TF_METRIC_ABBREVIATIONS[metric_name])
                # else:
                #     try:
                #         res_metrics.append(tf_metrics.get(metric_name))
                #     except ValueError:
                #         raise ValueError(f"Unknown metric '{metric_name}'")

            elif callable(metric_name):
                res_metrics.append(metric_name)
            else:
                raise ValueError(f"Unsupported metric type: {metric_name}")
        return res_metrics

    class BestModelMemory(Callback):
        def __init__(self):
            super(BestModelMemory, self).__init__()
            self.best_weights = None
            self.best_val_loss = np.inf

        def on_epoch_end(self, epoch, logs=None):
            val_loss = logs.get('val_loss')
            if val_loss < self.best_val_loss:
                self.best_val_loss = val_loss
                self.best_weights = self.model.get_weights()

        def on_train_end(self, logs=None):
            if self.best_weights is not None:
                self.model.set_weights(self.best_weights)

    class TFModelManager(BaseModelManager):
        def __init__(self, models=None, model_config=None):
            super(TFModelManager, self).__init__(models, model_config)
            self.framework = 'tensorflow'
        
        def train(self, dataset, training_params, input_dim=None, metrics=None, no_folds=False):
            # models = [self.model]
            models = [self.models[0]] if no_folds else self.models
            for (x_train, y_train, x_val, y_val), model in zip(dataset.fold_data('union', no_folds), models):
                print(f"Training fold with shapes: {x_train.shape}, {y_train.shape}, {x_val.shape}, {y_val.shape}")
                loss = training_params.get('loss', 'mse')

                # --- ENFORCE CORRECT LOSS FOR TENSORFLOW CLASSIFICATION ---
                # Detect task and num_classes using prepare_y logic
                # (We need to know if this is classification and how many classes)
                task = training_params.get('task', None)
                # If not explicitly set, try to infer from loss/metrics
                if task is None:
                    task = detect_task_type(loss, metrics or [])
                # Use the same logic as prepare_y to count classes
                num_classes = len(np.unique(y_train)) if task == 'classification' else None
                # Only override for TensorFlow classification
                if task == 'classification':
                    if num_classes == 2:
                        enforced_loss = 'binary_crossentropy'
                    else:
                        enforced_loss = 'sparse_categorical_crossentropy'
                    if loss != enforced_loss:
                        print(f"[INFO] Overriding loss '{loss}' with '{enforced_loss}' for TensorFlow classification (num_classes={num_classes})")
                    loss = enforced_loss

                # Reshape labels for sparse categorical cross-entropy
                if loss == 'sparse_categorical_crossentropy':
                    y_train = y_train.reshape(-1)  # Flatten to (n_samples,)
                    y_val = y_val.reshape(-1)
                    y_train, y_val = np.array(y_train, dtype=np.int32), np.array(y_val, dtype=np.int32)

                print(loss, metrics)

                x_train = tf.convert_to_tensor(x_train)
                y_train = tf.convert_to_tensor(y_train)
                x_val = tf.convert_to_tensor(x_val)
                y_val = tf.convert_to_tensor(y_val)

                print("Training with shapes:", x_train.shape, y_train.shape, x_val.shape, y_val.shape)
                model.compile(optimizer=training_params.get('optimizer', 'adam'),
                              loss=loss,
                              metrics=metrics)

                callbacks = []
                if training_params.get('early_stopping', True):
                    callbacks.append(EarlyStopping(monitor='val_loss', patience=training_params.get('patience', 50)))

                if training_params.get('cyclic_lr', False):
                    base_lr = training_params.get('base_lr', 1e-4)
                    max_lr = training_params.get('max_lr', 1e-2)
                    step_size = training_params.get('step_size', 2000)

                    def cyclic_lr(epoch):
                        cycle = np.floor(1 + epoch / (2 * step_size))
                        x = np.abs(epoch / step_size - 2 * cycle + 1)
                        lr = base_lr + (max_lr - base_lr) * max(0, (1 - x))
                        return lr

                    callbacks.append(LearningRateScheduler(cyclic_lr))

                # Add BestModelMemory callback to keep best weights in memory
                callbacks.append(BestModelMemory())

                # model.summary()

                model.fit(x_train, y_train,
                          validation_data=(x_val, y_val),
                          epochs=training_params.get('epochs', 100),
                          batch_size=training_params.get('batch_size', 32),
                          callbacks=callbacks,
                          verbose=training_params.get('verbose', 1))

        def predict(self, dataset, task, return_all=False, no_folds=False, raw_class_output=False):
            y_pred = None
            if len(self.models) == 1 or no_folds:
                y_pred = self.models[0].predict(dataset.x_test_('union'))
                if task == 'classification' and not raw_class_output:
                    return np.argmax(y_pred, axis=1)
                else:
                    return y_pred
            else:
                y_preds = [model.predict(dataset.x_test_('union')) for model in self.models]
                if task == 'classification':
                    if return_all:
                        if raw_class_output:
                            return y_preds  # Return raw probabilities/logits for each fold
                        y_preds = [np.argmax(y_pred, axis=1) for y_pred in y_preds]
                        return y_preds
                    else:
                        y_preds = np.mean(y_preds, axis=0)
                        return np.argmax(y_preds, axis=1) if not raw_class_output else y_preds
                else:
                    return y_preds if return_all else np.mean(y_preds, axis=0)
            
        def save_model(self, path):
            os.makedirs(os.path.dirname(path), exist_ok=True)
            for i, model in enumerate(self.models):
                model.save(path + f'_{i}.keras')

        def load_model(self, path):
            if os.path.exists(path):
                self.models = []
                for i in range(len(os.listdir(path))):
                    if os.path.exists(path + f'model_{i}.keras'):
                        self.models.append(load_model(path + f'model_{i}.keras'))
            else:
                raise FileNotFoundError(f"Model file not found at: {path}")


class SklearnModelManager(BaseModelManager):
    def __init__(self, models, model_config):
        super(SklearnModelManager, self).__init__(models, model_config)
        self.framework = 'sklearn'
        
    def train(self, dataset, training_params, metrics=None, no_folds=False):
        i = 0
        models = [self.models[0]] if no_folds else self.models
        for (x_train, y_train, x_val, y_val), model in zip(dataset.fold_data('concat', no_folds), models):
            print(f"Training fold {i + 1}, with shapes:", x_train.shape, y_train.shape, x_val.shape, y_val.shape)
            i += 1
            if 'accuracy' in metrics:
                if y_train.ndim > 1:
                    y_train = y_train.ravel()
                    y_val = y_val.ravel()
            model.fit(x_train, y_train)
            # if x_val is not None and y_val is not None:
            #     y_pred = model.predict(x_val)
            #     metrics = BaseModelManager.evaluate(y_val, y_pred, training_params.get('metrics', ['mse', 'r2']))
            #     print(f"Validation Metrics: {metrics}")

    def predict(self, dataset, task=None, return_all=False, no_folds=False, raw_class_output=False):
        if len(self.models) == 1 or no_folds:
            return self.models[0].predict(dataset.x_test_('concat'))
        else:
            y_preds = [model.predict(dataset.x_test_('concat')) for model in self.models]
            if return_all:
                return y_preds
            
            return np.mean(y_preds, axis=0)

class ModelManagerFactory:
    
    @staticmethod
    def get_model_manager(model_config, dataset, task):
        
        models, framework = ModelBuilderFactory.build_models(model_config, dataset, task)
        print("Using framework:", framework)

        if framework == 'tensorflow' and TF_AVAILABLE:
            return TFModelManager(models, model_config)
        elif framework == 'sklearn':
            return SklearnModelManager(models, model_config)
        # elif framework == 'pytorch' and TORCH_AVAILABLE:
            # return TorchModelManager(model, callable_model, model_params, model_config)
        else:
            raise ValueError(f"Unsupported framework or framework not available in the environment: {framework}")


# if TORCH_AVAILABLE:
    # import torch
    # import torch.optim as optim
    # from torch.utils.data import DataLoader, TensorDataset

    # class TorchModelManager(BaseModelManager):
    #     def __init__(self, model=None, callable_model=None, model_params=None, model_config=None):
    #         super(TorchModelManager, self).__init__(model, callable_model, model_params, model_config)
    #         self.framework = 'pytorch'

    #     def train(self, dataset, training_params):
    #         aggregation_type = self.get_aggregation_type(training_params)
    #         X_train, y_train = dataset.processed().train_data(aggregation_type=aggregation_type)
    #         X_val, y_val = dataset.processed().test_data(aggregation_type=aggregation_type)
    #         input_dim = X_train.shape[1]
    #         if self.model is None:
    #             self.instantiate_model(input_dim)
    #         # PyTorch training code
    #         self.model.train()
    #         optimizer = optim.Adam(self.model.parameters(), lr=training_params.get('base_lr', 1e-4))
    #         criterion = torch.nn.MSELoss()
    #         batch_size = training_params.get('batch_size', 32)
    #         epochs = training_params.get('epochs', 100)

    #         print("Training with shapes:", X_train.shape, y_train.shape, X_val.shape, y_val.shape)

    #         train_dataset = TensorDataset(torch.Tensor(X_train), torch.Tensor(y_train))
    #         val_dataset = TensorDataset(torch.Tensor(X_val), torch.Tensor(y_val))
    #         train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    #         val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

    #         best_val_loss = float('inf')
    #         best_weights = None

    #         for epoch in range(epochs):
    #             self.model.train()
    #             for X_batch, y_batch in train_loader:
    #                 optimizer.zero_grad()
    #                 outputs = self.model(X_batch)
    #                 loss = criterion(outputs, y_batch)
    #                 loss.backward()
    #                 optimizer.step()

    #             # Validation phase
    #             self.model.eval()
    #             val_loss = 0.0
    #             with torch.no_grad():
    #                 for X_batch, y_batch in val_loader:
    #                     outputs = self.model(X_batch)
    #                     loss = criterion(outputs, y_batch)
    #                     val_loss += loss.item()
    #             val_loss /= len(val_loader)

    #             # Update best weights
    #             if val_loss < best_val_loss:
    #                 best_val_loss = val_loss
    #                 best_weights = self.model.state_dict()

    #             print(f"Epoch [{epoch+1}/{epochs}], Validation Loss: {val_loss:.4f}")

    #         # Load best weights
    #         if best_weights is not None:
    #             self.model.load_state_dict(best_weights)

    #     def predict(self, dataset):
    #         aggregation_type = self.get_aggregation_type()
    #         X = dataset.processed().x_test(aggregation_type=aggregation_type)
    #         self.model.eval()
    #         with torch.no_grad():
    #             return self.model(torch.Tensor(X)).numpy()



#folder.py

from sklearn.model_selection import KFold, RepeatedKFold, StratifiedKFold, RepeatedStratifiedKFold, ShuffleSplit, GroupKFold, StratifiedShuffleSplit, BaseCrossValidator, TimeSeriesSplit
from typing import Any, Dict
from typing import Any, Dict, Union

from ._splitter import SystematicCircularSplitter, KBinsStratifiedSplitter, KMeansSplitter, KennardStoneSplitter, SPXYSplitter
import importlib

def get_splitter(split_config: Dict[str, Any]) -> BaseCrossValidator:
    """
    Factory function to instantiate cross-validation splitters.
    
    Parameters
    ----------
    split_config : Dict[str, Any]
        Configuration dictionary with 'method' and 'params'.
    
    Returns
    -------
    splitter : BaseCrossValidator
        An instance of a splitter compatible with scikit-learn.
    """
    if isinstance(split_config, BaseCrossValidator):
        return split_config
    
    if isinstance(split_config, object) and hasattr(split_config, 'split'):
        return split_config
    
    method = split_config.get('method')
    params = split_config.get('params', {})

    if isinstance(method, str):
        # Mapping of known scikit-learn splitters
        sklearn_splitters = {
            'KFold': KFold,
            'RepeatedKFold': RepeatedKFold,
            'StratifiedKFold': StratifiedKFold,
            'RepeatedStratifiedKFold': RepeatedStratifiedKFold,
            'ShuffleSplit': ShuffleSplit,
            'StratifiedShuffleSplit': StratifiedShuffleSplit,
            'GroupKFold': GroupKFold,
            'TimeSeriesSplit': TimeSeriesSplit,
            # TODO Add other sklearn splitters
            'SystematicCircularSplitter': SystematicCircularSplitter,
            'KBinsStratifiedSplitter': KBinsStratifiedSplitter,
            'KMeansSplitter': KMeansSplitter,
            'KennardStoneSplitter': KennardStoneSplitter,
            'SPXYSplitter': SPXYSplitter,
        }

        if method in sklearn_splitters:
            splitter_class = sklearn_splitters[method]
            return splitter_class(**params)
        else:
            # Attempt to load a custom splitter via import path
            try:
                module_name, class_name = method.rsplit('.', 1)
                module = importlib.import_module(module_name)
                splitter_class = getattr(module, class_name)
                if issubclass(splitter_class, BaseCrossValidator):
                    return splitter_class(**params)
                else:
                    raise ValueError(f"The class {class_name} is not a subclass of BaseCrossValidator.")
            except (ImportError, AttributeError, ValueError) as e:
                raise ValueError(f"Invalid splitter method: {method}.") from e
            
    elif isinstance(method, type):
        # If method is a class, instantiate it
        if issubclass(method, BaseCrossValidator):
            return method(**params)
        else:
            raise ValueError("Provided class is not a subclass of BaseCrossValidator.")
    elif isinstance(method, BaseCrossValidator):
        # If method is an instance, return it directly
        return method
    else:
        raise ValueError(f"Invalid method type: {type(method)}. Must be a string, class, or instance.")


def run_splitter(config, dataset):
    """
    Run the splitter on the dataset.
    """
    if config is None:
        return dataset
    
    splitter = get_splitter(config)
    folds = list(splitter.split(dataset.x_train_('concat', disable_augmentation=True), dataset.y_train_(disable_augmentation=True), dataset.group_train))
    return folds



from . import csv_loader
from . import data_config_parser
from . import dataset_loader
from . import dataset
from . import schema

__all__ = [
    'csv_loader',
    'data_config_parser',
    'dataset_loader',
    'dataset',
    'schema'
]


import math
from typing import List, Tuple, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

# -----------------------------------------------------------------------------
#  Light‑weight Keras → PyTorch helpers
# -----------------------------------------------------------------------------

class DepthwiseConv1d(nn.Conv1d):
    """Depth‑wise 1‑D convolution (grouped conv where groups=in_channels)."""

    def __init__(self, in_channels: int, kernel_size: int, depth_multiplier: int = 1, **conv_kwargs):
        super().__init__(
            in_channels,
            in_channels * depth_multiplier,
            kernel_size,
            groups=in_channels,
            **conv_kwargs,
        )


class SeparableConv1d(nn.Module):
    """Separable 1‑D convolution = depth‑wise + point‑wise."""

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int, depth_multiplier: int = 1, **conv_kwargs):
        super().__init__()
        self.depthwise = DepthwiseConv1d(in_channels, kernel_size, depth_multiplier, **conv_kwargs)
        self.pointwise = nn.Conv1d(in_channels * depth_multiplier, out_channels, kernel_size=1)

    def forward(self, x):  # (N,C,L)
        return self.pointwise(self.depthwise(x))


class SpatialDropout1d(nn.Dropout2d):
    """TensorFlow‑like spatial dropout along the channel axis."""

    def forward(self, x):  # (N,C,L)
        x = x.unsqueeze(3)  # (N,C,L,1)
        x = super().forward(x)
        return x.squeeze(3)


class GlobalAveragePooling1d(nn.Module):
    def forward(self, x):  # (N,C,L)
        return x.mean(dim=2)


# -----------------------------------------------------------------------------
#  Generic building blocks used by all architectures
# -----------------------------------------------------------------------------

def _activation(name: str):
    if name == "swish":
        return nn.SiLU()
    elif name == "selu":
        return nn.SELU()
    elif name == "elu":
        return nn.ELU()
    else:
        return nn.ReLU() if name == "relu" else getattr(nn, name.capitalize())()


def _norm(method: str, num_features: int):
    if method == "LayerNormalization":
        return nn.LayerNorm(num_features)
    return nn.BatchNorm1d(num_features)


# -----------------------------------------------------------------------------
#  Primitive model builders (DECON, NICON, Transformer)
# -----------------------------------------------------------------------------


def _build_decon(input_shape: Tuple[int, int], params: Dict, *, num_classes: int = 1):
    """Port of the TensorFlow `decon` family to PyTorch.

    This covers both regression (num_classes==1) and classification
    (num_classes>=2, including binary with sigmoid).
    """
    c, _ = input_shape  # channels, length
    layers: List[nn.Module] = [SpatialDropout1d(params.get("spatial_dropout", 0.2))]

    def _dconv(kernel, depth_multiplier):
        return DepthwiseConv1d(c, kernel, depth_multiplier=depth_multiplier, padding="same")

    # First block (7‑7)
    layers += [
        _dconv(7, 2), _activation("relu"),
        _dconv(7, 2), _activation("relu"),
        nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 2),
    ]

    # Second block (5‑5)
    layers += [
        _dconv(5, 2), _activation("relu"),
        _dconv(5, 2), _activation("relu"),
        nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 4),
    ]

    # Third block (9‑9)
    layers += [
        _dconv(9, 2), _activation("relu"),
        _dconv(9, 2), _activation("relu"),
        nn.MaxPool1d(2, 2), nn.BatchNorm1d(c * 8),
    ]

    # Separable + ordinary conv
    layers += [
        SeparableConv1d(c * 8, 64, 3, depth_multiplier=1, padding="same"), _activation("relu"),
        nn.Conv1d(64, 32, 3, padding="same"), _activation("relu"),
        nn.MaxPool1d(5, 3), SpatialDropout1d(0.1),
        nn.Flatten(),
        nn.Linear(32 * math.ceil(input_shape[1] / 8 / 3), 128), _activation("relu"),
        nn.Linear(128, 32), _activation("relu"),
        nn.Dropout(0.2),
    ]

    # Output
    if num_classes == 1:
        layers.append(nn.Linear(32, 1))
    elif num_classes == 2:
        layers.append(nn.Linear(32, 1))
        layers.append(nn.Sigmoid())
    else:
        layers.append(nn.Linear(32, num_classes))
        layers.append(nn.Softmax(dim=1))

    return nn.Sequential(*layers)


def _build_transformer(input_shape: Tuple[int, int], params: Dict, *, num_classes: int = 1):
    """1‑D Transformer with `nn.TransformerEncoder` blocks."""

    c, seq_len = input_shape
    head_size = params.get("head_size", 16)
    num_heads = params.get("num_heads", 2)
    ff_dim = params.get("ff_dim", 8)
    num_blocks = params.get("num_transformer_blocks", 1)
    dropout = params.get("dropout", 0.05)

    d_model = head_size * num_heads

    # Project the input channels to d_model, keeping length dimension
    proj_in = nn.Conv1d(c, d_model, 1)

    # Transformer encoder expects (L, N, E)
    encoder_layer = nn.TransformerEncoderLayer(
        d_model=d_model,
        nhead=num_heads,
        dim_feedforward=ff_dim * d_model,
        batch_first=False,
        dropout=dropout,
        activation="relu",
    )
    encoder = nn.TransformerEncoder(encoder_layer, num_layers=num_blocks)

    pooling = GlobalAveragePooling1d()

    mlp_units = params.get("mlp_units", [32, 8])
    mlp_dropout = params.get("mlp_dropout", 0.1)
    mlp: List[nn.Module] = []
    in_size = d_model
    for dim in mlp_units:
        mlp += [nn.Linear(in_size, dim), _activation("relu"), nn.Dropout(mlp_dropout)]
        in_size = dim

    # Output layer
    if num_classes == 1:
        out_layer = nn.Linear(in_size, 1)
    elif num_classes == 2:
        out_layer = nn.Sequential(nn.Linear(in_size, 1), nn.Sigmoid())
    else:
        out_layer = nn.Sequential(nn.Linear(in_size, num_classes), nn.Softmax(dim=1))

    class _Transformer(nn.Module):
        def __init__(self):
            super().__init__()
            self.proj_in = proj_in
            self.encoder = encoder
            self.pool = pooling
            self.mlp = nn.Sequential(*mlp)
            self.out = out_layer

        def forward(self, x):  # x: (N, C, L)
            z = self.proj_in(x)  # (N, d_model, L)
            z = z.permute(2, 0, 1)  # (L, N, d_model)
            z = self.encoder(z)
            z = z.permute(1, 2, 0)  # (N, d_model, L)
            z = self.pool(z)  # (N, d_model)
            z = self.mlp(z)
            return self.out(z)

    return _Transformer()


# -----------------------------------------------------------------------------
#  Public API — function names must match the original tensorflow file
# -----------------------------------------------------------------------------

from nirs4all.core.utils import framework  # noqa: E402, pylint: disable=C0413

# --- DECON family ------------------------------------------------------------

@framework("pytorch")
def decon(input_shape, params={}):
    return _build_decon(input_shape, params, num_classes=1)


@framework("pytorch")
def decon_classification(input_shape, num_classes=2, params={}):
    return _build_decon(input_shape, params, num_classes=num_classes)


@framework("pytorch")
def decon_Sep(input_shape, params={}):
    # Alias to standard decon for now – separable path is handled internally
    return _build_decon(input_shape, params, num_classes=1)


@framework("pytorch")
def decon_Sep_classification(input_shape, num_classes=2, params={}):
    return _build_decon(input_shape, params, num_classes=num_classes)


# --- Transformer family ------------------------------------------------------

@framework("pytorch")
def transformer(input_shape, params={}):
    return _build_transformer(input_shape, params, num_classes=1)


@framework("pytorch")
def transformer_VG(input_shape, params={}):
    # Passthrough with different defaults handled in params already
    return _build_transformer(input_shape, params, num_classes=1)


@framework("pytorch")
def transformer_classification(input_shape, num_classes=2, params={}):
    return _build_transformer(input_shape, params, num_classes=num_classes)


@framework("pytorch")
def transformer_VG_classification(input_shape, num_classes=2, params={}):
    return _build_transformer(input_shape, params, num_classes=num_classes)


# --- NICON family (simplified) ----------------------------------------------

def _build_nicon(input_shape, params, *, num_classes: int = 1):
    c, _ = input_shape
    layers: List[nn.Module] = [SpatialDropout1d(params.get("spatial_dropout", 0.08))]
    layers += [
        nn.Conv1d(c, params.get("filters1", 8), kernel_size=15, stride=5), _activation("selu"),
        nn.Dropout(params.get("dropout_rate", 0.2)),
        nn.Conv1d(params.get("filters1", 8), params.get("filters2", 64), kernel_size=21, stride=3), _activation("relu"),
        nn.BatchNorm1d(params.get("filters2", 64)),
        nn.Conv1d(params.get("filters2", 64), params.get("filters3", 32), kernel_size=5, stride=3), _activation("elu"),
        nn.BatchNorm1d(params.get("filters3", 32)),
        nn.Flatten(),
        nn.Linear(params.get("filters3", 32), params.get("dense_units", 16)), _activation("sigmoid"),
    ]

    if num_classes == 1:
        layers.append(nn.Linear(params.get("dense_units", 16), 1))
    elif num_classes == 2:
        layers += [nn.Linear(params.get("dense_units", 16), 1), nn.Sigmoid()]
    else:
        layers += [nn.Linear(params.get("dense_units", 16), num_classes), nn.Softmax(dim=1)]

    return nn.Sequential(*layers)


@framework("pytorch")
def nicon(input_shape, params={}):
    return _build_nicon(input_shape, params, num_classes=1)


@framework("pytorch")
def nicon_classification(input_shape, num_classes=2, params={}):
    return _build_nicon(input_shape, params, num_classes=num_classes)


@framework("pytorch")
def nicon_VG(input_shape, params={}):
    return _build_nicon(input_shape, params, num_classes=1)


@framework("pytorch")
def nicon_VG_classification(input_shape, num_classes=2, params={}):
    return _build_nicon(input_shape, params, num_classes=num_classes)


# --- Customizable variants ---------------------------------------------------

@framework("pytorch")
def customizable_nicon(input_shape, params={}):
    return _build_nicon(input_shape, params, num_classes=1)


@framework("pytorch")
def customizable_nicon_classification(input_shape, num_classes=2, params={}):
    return _build_nicon(input_shape, params, num_classes=num_classes)


@framework("pytorch")
def customizable_decon(input_shape, params={}):
    return _build_decon(input_shape, params, num_classes=1)


@framework("pytorch")
def decon_layer_classification(input_shape, num_classes=2, params={}):
    return _build_decon(input_shape, params, num_classes=num_classes)



# Empty __init__.py file to avoid circular imports
# Modules will be imported directly when needed rather than through the core package

__all__ = [
    'config',
    'finetuner',
    'manager',
    'model_manager',
    'processor',
    'runner',
    'utils'
]



"""
Manager module for handling experiment workflows and results.
"""

from nirs4all.core.manager.experiment_manager import *

__all__ = ["ExperimentManager"]  # Update this list with actual classes from the module


from ._splitter import CustomSplitter, KBinsStratifiedSplitter, KMeansSplitter, SystematicCircularSplitter, KennardStoneSplitter, SPXYSplitter
from ._folder import get_splitter, run_splitter
from ._helpers import train_test_split_idx


__all__ = [
    'CustomSplitter',
    'KBinsStratifiedSplitter',
    'KMeansSplitter',
    'SystematicCircularSplitter',
    'KennardStoneSplitter',
    'SPXYSplitter',
    'get_splitter',
    'run_splitter',
    'train_test_split_idx',
]



# processor.py

import numpy as np
from sklearn.pipeline import Pipeline
from joblib import Parallel, delayed
from sklearn.base import TransformerMixin, clone
import importlib
import inspect
import hashlib
from typing import Any, List, Tuple, Optional, Dict
import logging
import zlib

from ..data.dataset import Dataset
from ..data_splitters import run_splitter
from .model.model_builder_factory import ModelBuilderFactory
from .model.model_manager import ModelManagerFactory

def instantiate_class(class_name: str, params: dict) -> Any:
    """
    Instantiate a class given its fully qualified name and parameters.

    Args:
        class_name (str): The fully qualified class name (e.g., 'module.submodule.ClassName').
        params (dict): A dictionary of parameters to pass to the class constructor.

    Returns:
        Any: An instance of the specified class.

    Raises:
        ValueError: If the class cannot be found or instantiated.
    """
    try:
        components = class_name.split('.')
        module_name = '.'.join(components[:-1])
        class_name_only = components[-1]

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name_only)
        return class_(**params)
    except Exception as e:
        raise ValueError(f"Class {class_name} not found: {str(e)}") from e


def get_transformer(config: Any) -> TransformerMixin:
    """
    Get a transformer instance from the config.

    Args:
        config (Any): Configuration for the transformer, can be an instance, a class, a string,
                      a tuple, or a dict.

    Returns:
        TransformerMixin: An instance of a transformer.

    Raises:
        ValueError: If the configuration is invalid.
    """
    if isinstance(config, TransformerMixin):
        return config
    elif isinstance(config, tuple):
        class_def, params = config
        if isinstance(class_def, str):
            return instantiate_class(class_def, params)
        elif inspect.isclass(class_def):
            return class_def(**params)
        else:
            raise ValueError("Invalid transformer configuration in tuple.")
    elif isinstance(config, str):
        return instantiate_class(config, {})
    elif isinstance(config, dict):
        return instantiate_class(config['class'], config.get('params', {}))
    elif inspect.isclass(config):
        return config()
    else:
        raise ValueError(f"Invalid transformer configuration: {config}")


def run_pipeline(
    dataset: Dataset,
    x_pipeline: Any,
    y_pipeline: Any,
    logger: Any,
    cache: Optional[Dict[str, np.ndarray]] = None
) -> Dataset:
    """
    Run the pipeline on the dataset.

    Args:
        dataset (Dataset): The dataset to process.
        x_pipeline (Any): The pipeline configuration for x data.
        y_pipeline (Any): The pipeline configuration for y data.
        logger (Any): Logger object for logging.

    Returns:
        Dataset: The processed dataset.
    """
    dataset = exec_step_x(dataset, x_pipeline, logger, "", fit_test=False, cache=cache)
    dataset = exec_step_y(dataset, y_pipeline, fit_test=False)
    return dataset


def exec_step_y(
    dataset: Dataset,
    step: Any,
    fit_test: bool = False
) -> Dataset:
    """
    Execute the pipeline step for y data.

    Args:
        dataset (Dataset): The dataset containing y data.
        step (Any): The pipeline step to execute.
        fit_test (bool, optional): Whether to fit on test data. Defaults to False.

    Returns:
        Dataset: The dataset with transformed y data.

    Raises:
        NotImplementedError: If step is a dict or list, as augmentation is not authorized for y data.
        ValueError: If step format is invalid.
    """
    train_data, test_data = dataset.raw_y_train, dataset.raw_y_test
    if isinstance(step, dict):
        raise NotImplementedError("Sample and feature augmentation not authorized for y data.")
    elif isinstance(step, list):
        raise NotImplementedError("Sequential pipeline not authorized for y data.")
    elif isinstance(step, TransformerMixin) or isinstance(step, (str, tuple, dict)):
        transformer = get_transformer(step)
        if not fit_test:
            train_data = train_data if transformer is None else transformer.fit_transform(train_data)
            test_data = test_data if transformer is None else transformer.transform(test_data)
        else:
            combined_data = np.concatenate((train_data, test_data), axis=0)
            transformer.fit(combined_data)
            train_data = transformer.transform(train_data)
            test_data = transformer.transform(test_data)

        dataset.y_train, dataset.y_test, dataset.y_transformer = train_data, test_data, transformer
        return dataset
    elif step is None:
        return dataset
    else:
        raise ValueError(f"Invalid step format for y data: {step}")


def exec_step_x(
    dataset: Dataset,
    step: Any,
    logger: Any,
    indent: str,
    fit_test: bool = False,
    skip_test: bool = False,
    cache: Optional[Dict[str, np.ndarray]] = None
) -> Dataset:
    """
    Execute the pipeline step for x data.

    Args:
        dataset (Dataset): The dataset containing x data.
        step (Any): The pipeline step to execute.
        logger (Any): Logger object for logging.
        indent (str): Indentation string for logging.
        fit_test (bool, optional): Whether to fit on test data. Defaults to False.
        skip_test (bool, optional): Whether to skip processing test data. Defaults to False.
        cache (Optional[Dict[str, np.ndarray]], optional): Cache for storing transformed data. Defaults to None.

    Returns:
        Dataset: The dataset with transformed x data.

    Raises:
        ValueError: If the step format is invalid.
    """
    if cache is None:
        cache = {}

    train_data, test_data = dataset.raw_x_train, dataset.raw_x_test
    n_augmentations, n_samples, n_transformations, n_features = train_data.shape

    if isinstance(step, dict):
        sample_key = next((k for k in step if k in ["s", "samples", "sample_augmentation", "S"]), None)
        feature_key = next((k for k in step if k in ["f", "features", "feature_augmentation", "F", "parallel"]), None)
        split_key = next((k for k in step if k in ["split", "spl", "splitter", "SPL"]), None)

        if sample_key is not None:
            balanced = step.get("balance", False)
            sample_transformers = step[sample_key]

            if balanced:
                # Balanced augmentation
                y_train = dataset.raw_y_train
                classes, class_counts = np.unique(y_train, return_counts=True)
                max_count = np.max(class_counts)
                augmented_data_list = []
                augmented_labels_list = []

                for cls, count in zip(classes, class_counts):
                    num_needed = max_count - count
                    if num_needed > 0:
                        # Get indices of samples belonging to this class
                        class_indices = np.where(y_train.flatten() == cls)[0]
                        num_class_samples = len(class_indices)
                        repeats_per_sample = int(np.ceil(num_needed / num_class_samples))

                        for idx in class_indices:
                            sample_data = dataset.raw_x_train[:, idx:idx+1, :, :]
                            sample_label = y_train[idx:idx+1]

                            for _ in range(repeats_per_sample):
                                for st in sample_transformers:
                                    augmented_dataset = Dataset(_x_train=sample_data)
                                    augmented_dataset = exec_step_x(
                                        augmented_dataset, st, logger, indent + " ", fit_test, True, cache
                                    )
                                    augmented_data_list.append(augmented_dataset.raw_x_train)
                                    augmented_labels_list.append(sample_label)
                                    num_needed -= 1
                                    if num_needed <= 0:
                                        break
                                if num_needed <= 0:
                                    break
                            if num_needed <= 0:
                                break

                #print shape and number of element per class ({class: number of elements})
                print(dataset.raw_x_train.shape, dataset.raw_y_train.shape)
                print({cls: len(np.where(y_train.flatten() == cls)[0]) for cls in classes})
                print({cls: len(np.where(np.concatenate(augmented_labels_list).flatten() == cls)[0]) for cls in classes})
                if augmented_data_list:
                    # Concatenate augmented data and labels
                    augmented_data = np.concatenate(augmented_data_list, axis=1)
                    augmented_labels = np.concatenate(augmented_labels_list, axis=0)
                    # Update dataset with augmented data
                    dataset.raw_x_train = np.concatenate([dataset.raw_x_train, augmented_data], axis=1)
                    dataset._y_train = np.concatenate([dataset.raw_y_train, augmented_labels], axis=0)
                print(dataset.raw_x_train.shape, dataset.raw_y_train.shape)
                print({cls: len(np.where(dataset.y_train_().flatten() == cls)[0]) for cls in classes})
                # print({cls: len(np.where(np.concatenate(augmented_labels_list).flatten() == cls)[0]) for cls in classes})    
                
                return dataset
            else:
                # Original behavior (uniform augmentation)
                num_transformers = len(sample_transformers)
                train_data = np.repeat(train_data, num_transformers, axis=0)

                def process_sample_transformer(i_st):
                    i, st = i_st
                    aug_range = range(i * n_augmentations, (i + 1) * n_augmentations)
                    aug_data = train_data[aug_range, :, :, :]
                    augmented_dataset = Dataset(_x_train=aug_data)
                    augmented_dataset = exec_step_x(
                        augmented_dataset, st, logger, indent + " ", fit_test, True, cache
                    )
                    return (aug_range, augmented_dataset.raw_x_train)

                results = [process_sample_transformer((i, st)) for i, st in enumerate(sample_transformers)]

                for aug_range, aug_train_data in results:
                    train_data[aug_range, :, :, :] = aug_train_data

                dataset.raw_x_train = train_data
                return dataset

        elif feature_key is not None:
            # Feature augmentation
            # logger.info(indent + "Feature augmentation")
            feature_transformers = step[feature_key]
            num_transformers = len(feature_transformers)
            train_data = np.repeat(train_data, num_transformers, axis=2)
            if not skip_test:
                test_data = np.repeat(test_data, num_transformers, axis=2)

            def process_feature_transformer(i_ft):
                # logger.info(indent + f"Processing feature transformer {i_ft}")
                i, ft = i_ft
                tr_range = range(i * n_transformations, (i + 1) * n_transformations)
                tr_train_data = train_data[:, :, tr_range, :]
                tr_test_data = test_data[:, :, tr_range, :] if not skip_test else None

                transformed_dataset = Dataset(_x_train=tr_train_data, _x_test=tr_test_data)
                transformed_dataset = exec_step_x(
                    transformed_dataset, ft, logger, indent + " ", fit_test, skip_test, cache
                )
                return (tr_range, transformed_dataset.raw_x_train, transformed_dataset.raw_x_test)

            # Parallel execution
            # results = Parallel(n_jobs=-1, backend='loky')(
                # delayed(process_feature_transformer)((i, ft)) for i, ft in enumerate(feature_transformers)
            # )
            # Sequential execution
            results = [process_feature_transformer((i, ft)) for i, ft in enumerate(feature_transformers)]

            for tr_range, tr_train_data, tr_test_data in results:
                train_data[:, :, tr_range, :] = tr_train_data
                if not skip_test and tr_test_data is not None:
                    test_data[:, :, tr_range, :] = tr_test_data

            dataset.raw_x_train = train_data
            if not skip_test:
                dataset.raw_x_test = test_data
            return dataset

        elif split_key is not None:
            # Splitting dataset
            # logger.info(indent + "Splitting dataset")
            splitter_config = step[split_key]
            dataset.folds = run_splitter(splitter_config, dataset)
            return dataset

        elif "class" in step:
            transformer = get_transformer(step)
            return apply_step_x(dataset, transformer, logger, indent, fit_test, skip_test, cache)

        else:
            raise ValueError(f"Invalid step format in dict: {step}")

    elif isinstance(step, list):
        # Sequential pipeline
        # logger.info(indent + "Sequential pipeline")
        for s in step:
            dataset = exec_step_x(dataset, s, logger, indent + " ", fit_test, skip_test, cache)
        return dataset

    elif isinstance(step, TransformerMixin) or isinstance(step, (str, tuple, dict)):
        transformer = get_transformer(step)
        return apply_step_x(dataset, transformer, logger, indent, fit_test, skip_test, cache)

    elif inspect.isclass(step):
        transformer = get_transformer(step)
        return apply_step_x(dataset, transformer, logger, indent, fit_test, skip_test, cache)

    elif step is None:
        # Identity transformation
        # logger.info(indent + "Identity transformation")
        return dataset

    else:
        raise ValueError(f"Invalid step format: {step}")


def apply_step_x(
    dataset: Dataset,
    transformer: TransformerMixin,
    logger: Any,
    indent: str,
    fit_test: bool = False,
    skip_test: bool = False,
    cache: Optional[Dict[str, np.ndarray]] = None
) -> Dataset:
    """
    Apply a single transformer to the dataset.

    Args:
        dataset (Dataset): The dataset containing x data.
        transformer (TransformerMixin): The transformer to apply.
        logger (Any): Logger object for logging.
        indent (str): Indentation string for logging.
        fit_test (bool, optional): Whether to fit on test data. Defaults to False.
        skip_test (bool, optional): Whether to skip processing test data. Defaults to False.
        cache (Optional[Dict[str, np.ndarray]], optional): Cache for storing transformed data. Defaults to None.

    Returns:
        Dataset: The dataset with transformed x data.
    """
    logger = logging.getLogger(__name__)  # Ensure logging inside parallel workers
    logger.info(f"{indent}Processing ____")

    if cache is None:
        cache = {}
    
    train_data, test_data = dataset.raw_x_train, dataset.raw_x_test
    n_augmentations, n_samples, n_transformations, n_features = train_data.shape

    logger.info(indent + f"Applying transformer: {transformer.__class__.__name__}")

    # Generate transformer ID
    transformer_id = hashlib.md5(str(transformer).encode()).hexdigest()

    # Apply the transformer in parallel over transformations
    def process_transformation(transformation_index):
        logger = logging.getLogger('ExperimentRunner')
    
        train_slice = train_data[:, :, transformation_index, :]
        test_slice = test_data[:, :, transformation_index, :] if not skip_test else None
        
        # Clone the transformer to avoid parallel fitting conflicts
        # local_transformer = clone(transformer)
        local_transformer = transformer
        
        # Generate data IDs
        # train_data_id = hashlib.md5(train_slice.tobytes()).hexdigest()
        train_data_id = hash_data_crc32(train_slice.tobytes())
        # test_data_id = hashlib.md5(test_slice.tobytes()).hexdigest() if test_slice is not None else None
        test_data_id = hash_data_crc32(test_slice.tobytes()) if test_slice is not None else None
        
        # logger.info(indent + f"Processing transformation {transformation_index}: {transformer.__class__.__name__}")
        # logger.info(indent + f"Train data ID: {train_data_id}, Test data ID: {test_data_id}, cache keys: {cache.keys()}")

        # Check cache for transformed data
        cache_key_train = f"{train_data_id}_{transformer_id}"
        cache_key_test = f"{test_data_id}_{transformer_id}" if test_data_id else None

        if cache_key_train in cache:
            transformed_train = cache[cache_key_train]
            # logger.info(indent + f"Cache hit for train data at transformation {transformation_index}: {transformer.__class__.__name__}")
        else:
            # Reshape data for transformer
            train_data_view = train_slice.reshape(-1, n_features)
            if not fit_test:
                local_transformer.fit(train_data_view)
                transformed_train = local_transformer.transform(train_data_view)
            else:
                combined_data = train_data_view
                if test_slice is not None:
                    test_data_view = test_slice.reshape(-1, n_features)
                    combined_data = np.concatenate((train_data_view, test_data_view), axis=0)
                local_transformer.fit(combined_data)
                transformed_train = local_transformer.transform(train_data_view)
            transformed_train = transformed_train.reshape(n_augmentations, n_samples, n_features)
            cache[cache_key_train] = transformed_train

        if not skip_test and test_slice is not None:
            if cache_key_test in cache:
                transformed_test = cache[cache_key_test]
                # logger.info(indent + f"Cache hit for test data at transformation {transformation_index}")
            else:
                test_data_view = test_slice.reshape(-1, n_features)
                transformed_test = local_transformer.transform(test_data_view)
                transformed_test = transformed_test.reshape(test_slice.shape)
                cache[cache_key_test] = transformed_test
        else:
            transformed_test = None

        return (transformation_index, transformed_train, transformed_test)

    # Parallel execution
    # results = Parallel(n_jobs=4)(
        # delayed(process_transformation)(transformation_index) for transformation_index in range(n_transformations)
    # )
    # Sequential execution
    results = [process_transformation(transformation_index) for transformation_index in range(n_transformations)]

    for transformation_index, transformed_train, transformed_test in results:
        train_data[:, :, transformation_index, :] = transformed_train
        if not skip_test and transformed_test is not None:
            test_data[:, :, transformation_index, :] = transformed_test

    dataset.raw_x_train = train_data
    if not skip_test:
        dataset.raw_x_test = test_data
    return dataset


def hash_data_crc32(data):
    return zlib.crc32(data) & 0xffffffff  # Mask to ensure 32-bit output


class Processor:
    """
    Data processor class that handles the entire data pipeline.
    """
    
    def __init__(self, config=None, logger=None):
        """
        Initialize the processor with configuration.
        
        Args:
            config (dict): Configuration for the processor.
            logger: Logger instance.
        """
        self.config = config or {}
        self.logger = logger
        self.cache = {}
    
    def process_dataset(self, dataset):
        """
        Process the dataset through the pipeline.
        
        Args:
            dataset (Dataset): The dataset to process.
            
        Returns:
            Dataset: The processed dataset.
        """
        x_pipeline = self.config.get('x_pipeline')
        y_pipeline = self.config.get('y_pipeline')
        
        # Save categorical information before processing
        y_train_categorical_info = dataset.y_train_categorical_info
        y_test_categorical_info = dataset.y_test_categorical_info
        
        # Process data through pipeline
        processed_dataset = run_pipeline(dataset, x_pipeline, y_pipeline, self.logger, self.cache)
        
        # Restore categorical information after processing
        if y_train_categorical_info:
            processed_dataset.y_train_categorical_info = y_train_categorical_info
            
        if y_test_categorical_info:
            processed_dataset.y_test_categorical_info = y_test_categorical_info
            
        return processed_dataset
    
    def inverse_transform_predictions(self, y_pred, dataset):
        """
        Convert predictions back to original format, including categorical.
        
        Args:
            y_pred (numpy.ndarray): Predictions to transform
            dataset (Dataset): Dataset with transformation information
            
        Returns:
            numpy.ndarray: Transformed predictions
        """
        return dataset.inverse_transform(y_pred)




# base_finetuner.py
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
import optuna
from abc import ABC, abstractmethod
from ..model.model_builder_factory import ModelBuilderFactory
from ..model.model_manager import ModelManagerFactory
from ..utils import TF_AVAILABLE, TORCH_AVAILABLE
import copy
import numpy as np


class BaseFineTuner(ABC):
    def __init__(self, model_manager):
        self.model_manager = model_manager

    @abstractmethod
    def finetune(self, dataset, finetune_params, src_training_params=None, metrics=None, task=None):
        pass


class OptunaFineTuner(BaseFineTuner):
    def __init__(self, model_manager):
        super().__init__(model_manager)
        self.model_config = copy.deepcopy(model_manager.model_config)  # Access initial model_config

    def finetune(self, dataset, finetune_params, src_training_params=None, metrics=None, task=None):
        union = 'concat' if self.model_manager.framework == 'sklearn' else 'union'
        model_config = copy.deepcopy(self.model_config)
        model = ModelBuilderFactory.build_single_model(model_config, dataset, task)
        finetune_model_manager = ModelManagerFactory.get_model_manager(model, dataset, task)  # TODO avoid multiple models if not necessary
        
        def objective(trial):
            # Deep copy to avoid mutations
            model_config = copy.deepcopy(self.model_config)
            model_params = {}
            training_params = {}

            # Extract model parameters
            for key, values in finetune_params.get('model_params', {}).items():
                if isinstance(values, list):
                    model_params[key] = trial.suggest_categorical(key, values)
                elif isinstance(values, tuple):
                    if values[0] == 'int':
                        model_params[key] = trial.suggest_int(key, values[1], values[2])
                    elif values[0] == 'float':
                        model_params[key] = trial.suggest_float(key, values[1], values[2])

            # Update the model_config with new parameters
            # if 'model_params' not in model_config:
                # model_config['model_params'] = {}
            # model_config['model_params'].update(model_params)

            # Similarly extract training parameters
            for key, values in finetune_params.get('training_params', {}).items():
                if isinstance(values, list):
                    training_params[key] = trial.suggest_categorical(key, values)
                elif isinstance(values, tuple):
                    if values[0] == 'int':
                        training_params[key] = trial.suggest_int(key, values[1], values[2])
                    elif values[0] == 'float':
                        training_params[key] = trial.suggest_float(key, values[1], values[2])

            # Build the model with updated parameters
            model = ModelBuilderFactory.build_single_model(model_config, dataset, task, model_params)
            finetune_model_manager.models = [model]
            
            # merge training params and src_training_params, keeping the training_params
            if src_training_params is not None:
                for key, value in src_training_params.items():
                    if key not in training_params:
                        training_params[key] = value
            
            # Train the model
            finetune_model_manager.train(dataset, training_params=training_params, metrics=metrics, no_folds=True)

            # Predict and evaluate
            y_pred = finetune_model_manager.predict(dataset, task, no_folds=True, raw_class_output=(task == 'classification'))
            y_true = dataset.y_test
            if task == 'classification':
                if y_pred.ndim > 1:
                    if y_pred.shape[-1] == dataset.num_classes and dataset.num_classes > 1:
                        y_pred = np.argmax(y_pred, axis=-1)
                    elif y_pred.shape[-1] == 1 or dataset.num_classes == 1:
                        y_pred = (y_pred >= 0.5).astype(int).flatten()
                    else:
                        raise ValueError("Unexpected output shape for classification task")
                else:
                    y_pred = (y_pred >= 0.5).astype(int)
            
            scores = finetune_model_manager.evaluate(y_true, y_pred, metrics)
            
            return scores[metrics[0]]

        study = optuna.create_study(direction="minimize")
        study.optimize(objective, n_trials=finetune_params.get('n_trials', 100), n_jobs=-1)

        best_params = study.best_params
        # print(f"Best hyperparameters: {best_params}")

        # Update the model_manager with the best model
        # best_trial = study.best_trial
        # Rebuild the model with best parameters
        best_model_params = {}
        best_training_params = {}
        for key, value in best_params.items():
            if key in finetune_params.get('model_params', {}):
                best_model_params[key] = value
            elif key in finetune_params.get('training_params', {}):
                best_training_params[key] = value
                
        print(f"Best model params: {best_model_params}")

        # self.model_manager.model_config['model_params'].update(best_model_params)
        # Build the best model
        best_models, _ = ModelBuilderFactory.build_models(self.model_manager.model_config, dataset, task, best_model_params)
        self.model_manager.models = best_models

        if src_training_params is not None:
            for key, value in src_training_params.items():
                if key not in best_training_params:
                    best_training_params[key] = value

        # Retrain best model using original metrics
        self.model_manager.train(dataset, training_params=best_training_params, metrics=metrics)

        return best_params


class SklearnFineTuner(BaseFineTuner):
    def __init__(self, model_manager):
        super().__init__(model_manager)
        self.model_config = copy.deepcopy(model_manager.model_config)

    def finetune(self, dataset, finetune_params, src_training_params=None, metrics=None, task=None):
        union = 'concat' if self.model_manager.framework == 'sklearn' else 'union'
        x_train, y_train = dataset.x_train_(union), dataset.y_train

        approach = finetune_params.get('approach', 'grid')
        param_dict = {}

        # Convert model_params for sklearn GridSearchCV or RandomizedSearchCV
        for key, values in finetune_params.get('model_params', {}).items():
            if isinstance(values, list):
                param_dict[key] = values  # Discrete choices
            elif isinstance(values, tuple):
                if values[0] == 'int':
                    # print("int", values[1], values[2], key)
                    param_dict[key] = list(range(values[1], values[2] + 1))
                elif values[0] == 'float':
                    param_dict[key] = np.linspace(values[1], values[2], num=5)

        n_trials = finetune_params.get('n_trials', 30)
        n_jobs = finetune_params.get('n_jobs', -1)

        # Build base estimator
        estimator = ModelBuilderFactory.build_single_model(self.model_config, dataset=dataset, task=task)

        # cv = finetune_params.get('cv', 5)
        
        # print(param_dict)
        if approach == 'grid':
            search = GridSearchCV(estimator, param_dict, cv=None, n_jobs=n_jobs)
        elif approach == 'random':
            search = RandomizedSearchCV(estimator, param_dict, n_iter=n_trials, cv=None, n_jobs=n_jobs)
        else:
            search = GridSearchCV(estimator, param_dict, cv=None, n_jobs=n_jobs)

        search.fit(x_train, y_train)
        best_params = search.best_params_
        best_models, _ = ModelBuilderFactory.build_models(search.best_estimator_, dataset, task, best_params)
        self.model_manager.models = best_models

        # Update the model_config
        print(f"Best model params: {best_params}")
        # self.model_manager.model_config['model_params'].update(best_params)

        # Optional training with new parameters
        training_params = finetune_params.get('training_params', {})
        
        if src_training_params is not None:
            for key, value in src_training_params.items():
                if key not in training_params:
                    training_params[key] = value
        self.model_manager.train(dataset, training_params=training_params, metrics=metrics)

        return best_params


class FineTunerFactory:
    @staticmethod
    def get_fine_tuner(tuner_type, model_manager):
        if tuner_type == 'optuna':
            return OptunaFineTuner(model_manager)
        # elif tuner_type == 'hyperband' and TF_AVAILABLE:
            # return HyperbandFineTuner(model_manager)
        elif tuner_type == 'sklearn':
            return SklearnFineTuner(model_manager)
        # elif tuner_type == 'torch' and TORCH_AVAILABLE:
            # return TorchFineTuner(model_manager)
        else:
            raise ValueError(f"Unsupported fine-tuner type or framework not available: {tuner_type}")


# if TF_AVAILABLE:
#     from keras_tuner import Hyperband

#     class HyperbandFineTuner(BaseFineTuner):
#         def __init__(self, model_manager):
#             super().__init__(model_manager)
#             self.model_config = copy.deepcopy(model_manager.model_config)

#         def finetune(self, dataset, finetune_params):
#             aggregation_type = self.model_manager.get_aggregation_type(finetune_params)
#             X_train, y_train = dataset.processed().train_data(aggregation_type=aggregation_type)
#             X_val, y_val = dataset.processed().test_data(aggregation_type=aggregation_type)

#             def build_model(hp):
#                 model_params = {}
#                 for key, values in finetune_params.get('model_params', {}).items():
#                     if isinstance(values, list):
#                         model_params[key] = hp.Choice(key, values)
#                     elif isinstance(values, tuple):
#                         if values[0] == 'int':
#                             model_params[key] = hp.Int(key, values[1], values[2])
#                         elif values[0] == 'float':
#                             model_params[key] = hp.Float(key, values[1], values[2])

#                 # Update the model_config with new parameters
#                 model_config = copy.deepcopy(self.model_config)
#                 if 'model_params' not in model_config:
#                     model_config['model_params'] = {}
#                 model_config['model_params'].update(model_params)

#                 model, _ = ModelBuilderFactory.build_model(model_config)
#                 model.compile(optimizer=finetune_params.get('optimizer', 'adam'),
#                               loss=finetune_params.get('loss', 'mse'),
#                               metrics=finetune_params.get('metrics', ['mse']))
#                 return model

#             tuner = Hyperband(
#                 build_model,
#                 objective="val_loss",
#                 max_epochs=finetune_params.get('max_epochs', 50),
#                 factor=finetune_params.get('factor', 3),
#                 directory=finetune_params.get('directory', 'hyperband_results'),
#                 project_name="hyperband_tuning"
#             )
#             tuner.search(X_train, y_train, validation_data=(X_val, y_val))
#             best_hps = tuner.get_best_hyperparameters(num_trials=1)[0]
#             # print(f"Best hyperparameters: {best_hps.values}")

#             # Update the model_manager with the best model
#             self.model_manager.model_config['model_params'].update(best_hps.values)
#             best_model = tuner.get_best_models(num_models=1)[0]
#             self.model_manager.model = best_model

#             # Optionally, retrain the best model on the full training data
#             training_params = finetune_params.get('training_params', {})
#             self.model_manager.train(dataset, training_params=training_params)

#             return best_hps.values


# if TORCH_AVAILABLE:
#     import torch
#     import torch.optim as optim
#     from torch.utils.data import DataLoader, TensorDataset

#     class TorchFineTuner(BaseFineTuner):
#         def __init__(self, model_manager):
#             super().__init__(model_manager)
#             self.model_config = copy.deepcopy(model_manager.model_config)

#         def finetune(self, dataset, finetune_params):
#             aggregation_type = self.model_manager.get_aggregation_type(finetune_params)
#             X_train, y_train = dataset.processed().train_data(aggregation_type=aggregation_type)
#             X_val, y_val = dataset.processed().test_data(aggregation_type=aggregation_type)

#             model_params_list = []
#             training_params_list = []

#             # Prepare grid of parameters
#             model_params_grid = {}
#             for key, values in finetune_params.get('model_params', {}).items():
#                 if isinstance(values, list):
#                     model_params_grid[key] = values
#                 elif isinstance(values, tuple):
#                     if values[0] == 'int':
#                         model_params_grid[key] = list(range(values[1], values[2] + 1))
#                     elif values[0] == 'float':
#                         model_params_grid[key] = np.linspace(values[1], values[2], num=5)

#             from itertools import product
#             param_keys = list(model_params_grid.keys())
#             param_values = list(model_params_grid.values())
#             param_combinations = [dict(zip(param_keys, v)) for v in product(*param_values)]

#             best_val_loss = float('inf')
#             best_params = None
#             best_model_state = None

#             for params in param_combinations:
#                 # Update model_config with current params
#                 model_config = copy.deepcopy(self.model_config)
#                 if 'model_params' not in model_config:
#                     model_config['model_params'] = {}
#                 model_config['model_params'].update(params)

#                 # Build the model
#                 model, _ = ModelBuilderFactory.build_model(model_config)
#                 # Create a new ModelManager instance with the updated model
#                 model_manager_class = type(self.model_manager)
#                 temp_model_manager = model_manager_class(model=model, model_config=model_config)
#                 temp_model_manager.framework = 'pytorch'

#                 # Train the model
#                 training_params = finetune_params.get('training_params', {})
#                 temp_model_manager.train(dataset, training_params=training_params)

#                 # Validate the model
#                 y_pred = temp_model_manager.predict(dataset)
#                 y_true = y_val
#                 metrics = temp_model_manager.evaluate(y_true, y_pred)
#                 val_loss = metrics['mse']

#                 if val_loss < best_val_loss:
#                     best_val_loss = val_loss
#                     best_params = params
#                     best_model_state = temp_model_manager.model.state_dict()

#             # Update the model_manager with the best model
#             self.model_manager.model_config['model_params'].update(best_params)
#             best_model, _ = ModelBuilderFactory.build_model(self.model_manager.model_config)
#             best_model.load_state_dict(best_model_state)
#             self.model_manager.model = best_model

#             # Optionally, retrain the best model on the full training data
#             self.model_manager.train(dataset, training_params=finetune_params.get('training_params', {}))

#             return best_params

# Example Configurations for Fine-Tuners
#
# Optuna Fine-Tuner Config
# finetune_params = {
#     'model_params': {
#         'n_components': ('int', 1, 10),
#         'hidden_layers': [1, 2, 3, 4],
#         'activation': ['relu', 'tanh', 'sigmoid']
#     },
#     'training_params': {
#         'epochs': ('int', 10, 100),
#         'batch_size': [16, 32, 64],
#         'learning_rate': ('float', 1e-5, 1e-2),
#         'optimizer': ['adam', 'sgd']
#     },
#     'n_trials': 20
# }
#
# Hyperband Fine-Tuner Config
# finetune_params = {
#     'model_params': {
#         'n_components': ('int', 2, 15),
#         'dropout_rate': ('float', 0.1, 0.5)
#     },
#     'training_params': {
#         'epochs': ('int', 5, 50),
#         'batch_size': [32, 64, 128],
#         'learning_rate': ('float', 1e-4, 1e-1)
#     },
#     'max_epochs': 50,
#     'factor': 3,
#     'directory': 'hyperband_results',
#     'project_name': 'hyperband_tuning'
# }
#
# Sklearn Fine-Tuner Config
# finetune_params = {
#     'model_params': {
#         'n_estimators': ('int', 50, 150),
#         'max_depth': [3, 5, 10],
#         'learning_rate': ('float', 0.01, 0.2)
#     },
#     'approach': 'grid',
#     'cv': 5,
#     'n_trials': 10,
#     'n_jobs': -1,
#     'training_params': {
#         'epochs': 100,
#         'batch_size': 32
#     }
# }
#
# Torch Fine-Tuner Config
# finetune_params = {
#     'model_params': {
#         'num_layers': [1, 2, 3],
#         'units_per_layer': ('int', 16, 128)
#     },
#     'training_params': {
#         'learning_rate': ('float', 1e-5, 1e-2),
#         'batch_size': [16, 32, 64],
#         'epochs': 20
#     }
# }



import numpy as np
import scipy
from scipy import signal
from scipy.ndimage import gaussian_filter1d
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils import check_array
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted


class Baseline(TransformerMixin, BaseEstimator):
    """
    Removes baseline (mean) from each spectrum.

    Parameters
    ----------
    copy : bool, optional
        Flag to indicate whether to make a copy of the object, by default True.
    """

    def __init__(self, *, copy=True):
        """
        Constructor for the Baseline class.

        Parameters
        ----------
        copy : bool, optional
            Flag to indicate whether to make a copy of the object, by default True.
        """

        self.copy = copy

    def _reset(self):
        if hasattr(self, "mean_"):
            del self.mean_

    def fit(self, X, y=None):
        """
        Compute the minimum and maximum to be used for later scaling.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data used to compute the per-feature minimum and maximum
            used for later scaling along the features axis.
        y : None
            Ignored.

        Returns
        -------
        self : object
            Fitted Baseline object.
        """

        self._reset()
        return self.partial_fit(X, y)

    def partial_fit(self, X, y=None):
        if scipy.sparse.issparse(X):
            raise TypeError("Baseline does not support scipy.sparse input")

        first_pass = not hasattr(self, "mean_")
        X = self._validate_data(X, reset=first_pass, dtype=FLOAT_DTYPES, estimator=self)

        self.mean_ = np.mean(X, axis=0)
        return self

    def transform(self, X, y=None):
        check_is_fitted(self)

        X = self._validate_data(
            X, reset=False, copy=self.copy, dtype=FLOAT_DTYPES, estimator=self
        )

        X = X - self.mean_
        return X

    def inverse_transform(self, X, y=None):
        check_is_fitted(self)

        X = check_array(X, copy=self.copy, dtype=FLOAT_DTYPES)

        X = X + self.mean_
        return X

    def _more_tags(self):
        return {"allow_nan": False}


def baseline(spectra):
    """
    Removes baseline (mean) from each spectrum.

    Parameters
    ----------
    spectra : numpy.ndarray
        NIRS data matrix.

    Returns
    -------
    numpy.ndarray
        Mean-centered NIRS data matrix.
    """

    return spectra - np.mean(spectra, axis=0)


def detrend(spectra, bp=0):
    """
    Perform spectral detrending to remove linear trend from data.

    Parameters
    ----------
    spectra : numpy.ndarray
        NIRS data matrix.
    bp : list, optional
        A sequence of break points. If given, an individual linear fit is performed for each part of data between two break points.
        Break points are specified as indices into data. Default is 0.

    Returns
    -------
    numpy.ndarray
        Detrended NIR spectra.
    """

    return signal.detrend(spectra, bp=bp)


class Detrend(TransformerMixin, BaseEstimator):
    """
    Perform spectral detrending to remove linear trend from data.

    Parameters
    ----------
    bp : int, optional
        Breakpoints for piecewise linear detrending. Default is 0.
    copy : bool, optional
        Whether to make a copy of the input data. Default is True.
    """

    def __init__(self, bp=0, *, copy=True):
        self.copy = copy
        self.bp = bp

    def _reset(self):
        """
        Reset internal data-dependent state of the transformer.
        """

        pass

    def fit(self, X, y=None):
        """
        Fit the transformer to the data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.
        y : None
            Ignored.

        Returns
        -------
        self : object
            Returns self.
        """

        if scipy.sparse.issparse(X):
            raise ValueError('Sparse matrices not supported!"')
        return self

    def transform(self, X, copy=None):
        """
        Transform the data by removing linear trend.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.
        copy : bool or None, optional
            Whether to make a copy of the input data. If None, `self.copy` is used. Default is None.

        Returns
        -------
        numpy.ndarray
            The transformed data.
        """

        if scipy.sparse.issparse(X):
            raise ValueError('Sparse matrices not supported!"')

        X = self._validate_data(
            X, reset=False, copy=self.copy, dtype=FLOAT_DTYPES, estimator=self
        )

        X = detrend(X, bp=self.bp)

        return X

    def _more_tags(self):
        """
        Get tags for the estimator.

        Returns
        -------
        dict
            Dictionary of tags for the estimator.
        """

        return {"allow_nan": False}


def gaussian(spectra, order=2, sigma=1):
    """
    Computes 1D gaussian filter using scipy.ndimage gaussian 1d filter.

    Parameters
    ----------
    spectra : numpy.ndarray
        NIRS data matrix.
    order : float, optional
        Order of the derivation.
    sigma : int, optional
        Sigma of the gaussian.

    Returns
    -------
    numpy.ndarray
        Gaussian NIR spectra.
    """

    return gaussian_filter1d(spectra, order=order, sigma=sigma)


class Gaussian(TransformerMixin, BaseEstimator):
    def __init__(self, order=2, sigma=1, *, copy=True):
        """
        Initialize Gaussian filter.

        Parameters
        ----------
        order : float, optional
            Order of the derivation.
        sigma : int, optional
            Sigma of the gaussian.
        copy : bool, default=True
            Whether to make a copy of the input data.
        """

        self.copy = copy
        self.order = order
        self.sigma = sigma

    def _reset(self):
        """
        Reset internal data-dependent state of the Gaussian filter.
        """

        pass

    def fit(self, X, y=None):
        """
        Fit the Gaussian filter.

        Parameters
        ----------
        X : numpy.ndarray
            Input data.
        y : None
            Ignored.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

        if scipy.sparse.issparse(X):
            raise ValueError("SavitzkyGolay does not support scipy.sparse input")
        return self

    def transform(self, X, copy=None):
        """
        Transform the input data using the Gaussian filter.

        Parameters
        ----------
        X : numpy.ndarray
            Input data.
        copy : bool, default=None
            Whether to make a copy of the input data.

        Returns
        -------
        numpy.ndarray
            Transformed data.
        """

        if scipy.sparse.issparse(X):
            raise ValueError('Sparse matrices not supported!"')

        X = self._validate_data(
            X, reset=False, copy=self.copy, dtype=np.float64, estimator=self
        )

        X = gaussian(X, order=self.order, sigma=self.sigma)

        return X

    def _more_tags(self):
        """
        Provide additional tags for the Gaussian filter.
        """

        return {"allow_nan": False}



# config.py
import itertools
import dataclasses
from typing import List, Optional, Union

@dataclasses.dataclass
class Config:
    dataset: Union[str, object]
    x_pipeline: Optional[Union[str, object]] = None
    y_pipeline: Optional[Union[str, object]] = None
    model: Optional[Union[str, object]] = None
    experiment: Optional[dict] = None
    seed: Optional[int] = None


    def validate(self, dataset_instance):
        experiment_config = self.experiment or {}
        finetune_params = experiment_config.get('finetune_params', {})
        training_params = experiment_config.get('training_params', {})
        action = experiment_config.get('action', 'train')
        
        task = experiment_config.get('task', None)
        metrics = experiment_config.get('metrics', None)
        loss = training_params.get('loss', None)
        classification_losses = {'binary_crossentropy', 'categorical_crossentropy', 'sparse_categorical_crossentropy'}
        classification_metrics = {'accuracy', 'acc', 'precision', 'recall', 'f1', 'auc'}
        
        if task is None:
            if loss is not None:
                if loss in classification_losses:
                    task = 'classification'
                    if metrics is None:
                        metrics = ['accuracy']
                else:
                    task = 'regression'
                    if metrics is None:
                        metrics = ['mse', 'mae']
            elif metrics is not None:
                if any(metric in classification_metrics for metric in metrics):
                    task = 'classification'
                    if dataset_instance.num_classes is None:
                        raise ValueError("Number of classes is not defined in dataset. Please specify the number of classes in the dataset config.")
                    elif dataset_instance.num_classes == 2:
                        training_params['loss'] = 'binary_crossentropy'
                    else:
                        training_params['loss'] = 'sparse_categorical_crossentropy'
                else:
                    task = 'regression'
                    training_params['loss'] = 'mse'
            else:
                task = 'regression'
                if metrics is None:
                    metrics = ['mse', 'mae']
        else:
            if task == 'classification' and metrics is None:
                metrics = ['accuracy']
            elif task == 'regression' and metrics is None:
                metrics = ['mse', 'mae']
                
            if task == 'classification' and 'loss' not in training_params:
                if dataset_instance.num_classes == 2:
                    training_params['loss'] = 'binary_crossentropy'
                else:
                    training_params['loss'] = 'sparse_categorical_crossentropy'
            elif task == 'regression' and 'loss' not in training_params:
                training_params['loss'] = 'mse'
            
        experiment_config['metrics'] = metrics
        experiment_config['task'] = task
        if task == 'classification':
            experiment_config['num_classes'] = dataset_instance.num_classes
        self.experiment = experiment_config
        
        return action, metrics, training_params, finetune_params, task





# @dataclasses.dataclass
# class Configs_Generator:
#     datasets: List[str]
#     model_experiments: List[dict]  # List of tuples (model_config, experiment)
#     preparations: Optional[List[str]] = None
#     scalers: Optional[List[str]] = None
#     augmenters: Optional[List[str]] = None
#     preprocessings: Optional[List[str]] = None
#     reporting: Optional[dict] = None
#     seeds: Optional[List[int]] = None

#     def generate_configs(self):
#         self.preparations = self.preparations or [None]
#         self.scalers = self.scalers or [None]
#         self.preprocessings = self.preprocessings or [None]

#         for dataset, (model_config, experiment), preparation, scaler, preprocessing, seed in itertools.product(
#             self.datasets, self.model_experiments, self.preparations, self.scalers, self.preprocessings, self.seeds
#         ):
#             yield Config(dataset, model_config, preparation, scaler, preprocessing, experiment, seed, self.reporting)



# runner.py

import os
import numpy as np

from ..data.dataset_loader import get_dataset
from .processor import run_pipeline
from nirs4all.core.finetuner.base_finetuner import FineTunerFactory
from .model.model_manager import ModelManagerFactory
from .manager.experiment_manager import ExperimentManager

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'


class ExperimentRunner:
    def __init__(self, configs, results_dir='results', resume_mode='restart', verbose=0):
        self.configs = configs
        self.manager = ExperimentManager(results_dir, resume_mode, verbose=3)
        self.logger = self.manager.logger
        self.results_dir = results_dir
        self.cache = {}

    def _run_config(self, config):
        self.cache = {}
        self.logger.info("=" * 80)
        self.logger.info("### LOADING DATASET ###")
        dataset = get_dataset(config.dataset)
        self.logger.info(dataset)

        self.logger.info("### PROCESSING DATASET ###")
        dataset = run_pipeline(dataset, config.x_pipeline, config.y_pipeline, self.logger, self.cache)
        self.logger.info(dataset)
        
        # len of unique classes for y_train merged with y_test
        dataset.num_classes = len(np.unique(np.concatenate([dataset.y_train_init, dataset.y_test_init])))
        
        action, metrics, training_params, finetune_params, task = config.validate(dataset)
        
        self.logger.info("### PREPARING MODEL ###")
        model_manager = None
        model_config = config.model
        if model_config is not None:
            model_manager = ModelManagerFactory.get_model_manager(model_config, dataset, task)
        
        self.logger.info("Running config > %s", self.manager.make_config_serializable(config))
        self.manager.prepare_experiment(config)
        
        preds, scores, best_params = None, None, None
        if model_manager is None:
            self.logger.warning("No model manager found. Skipping training/prediction.")
            return dataset, preds, scores, best_params

        if action == 'predict':
            preds, scores, best_params = self._predict(model_manager, dataset, metrics, task)
        elif action == 'train':
            preds, scores, best_params = self._train(model_manager, dataset, training_params, metrics, task)
        elif action == 'finetune':
            preds, scores, best_params = self._fine_tune(model_manager, dataset, finetune_params, training_params, metrics, task)
    
        return dataset, preds, scores, best_params

    def run(self):
        if not isinstance(self.configs, list):
            self.configs = [self.configs]

        datasets, predictions, scores, best_params = [], [], [], []

        for config in self.configs:
            self.logger.info("=" * 80)
            self.logger.info("Running config: %s", config)
            dataset_, preds_, scores_, best_params_ = self._run_config(config)

            datasets.append(dataset_)
            predictions.append(preds_)
            scores.append(scores_)
            best_params.append(best_params_)

        self.logger.info("All experiments completed.")
        return datasets, predictions, scores, best_params
    
    def _evaluate_and_save_results(self, model_manager, dataset, metrics, best_params=None, task=None):
        # Request raw outputs from the model
        y_pred_raw_outputs = model_manager.predict(dataset, task, return_all=True, raw_class_output=(task == 'classification'))
        
        # Get the initial true labels (these might be encoded)
        y_true_initial = dataset.y_test_init.ravel()  # Ensure y_true is 1D

        # Prepare y_true for evaluation and saving: Inverse transform if classification
        y_true_for_eval_and_save = y_true_initial
        if task == 'classification':
            if hasattr(dataset, 'inverse_transform') and callable(dataset.inverse_transform):
                y_true_for_eval_and_save = dataset.inverse_transform(y_true_initial)
            else:
                print(("[WARNING] Dataset does not have a callable 'inverse_transform' method. "
                       "y_true for classification might remain in its initial (potentially encoded) form, "
                       "which could affect metrics and saved results if predictions are in original label format."))
        
        # Ensure y_true_for_eval_and_save is consistently 1D array after potential inverse_transform
        if not isinstance(y_true_for_eval_and_save, np.ndarray):
            y_true_for_eval_and_save = np.array(y_true_for_eval_and_save)
        if y_true_for_eval_and_save.ndim > 1 and y_true_for_eval_and_save.shape[1] == 1:
            y_true_for_eval_and_save = y_true_for_eval_and_save.ravel()

        if isinstance(y_pred_raw_outputs, list):  # Handling multiple folds
            raw_preds_model_output_folds = []
            fold_scores = []
            all_preds_inverse_transformed_for_saving = []

            for y_pred_i_raw_output_fold in y_pred_raw_outputs:
                raw_preds_model_output_folds.append(y_pred_i_raw_output_fold)

                y_pred_class_fold = None
                if task == 'classification':
                    if y_pred_i_raw_output_fold.ndim > 1:
                        if y_pred_i_raw_output_fold.shape[-1] == dataset.num_classes and dataset.num_classes > 1:  # Multi-class
                            y_pred_class_fold = np.argmax(y_pred_i_raw_output_fold, axis=-1)
                        elif y_pred_i_raw_output_fold.shape[-1] == 1 or dataset.num_classes == 1:  # Binary
                            y_pred_class_fold = (y_pred_i_raw_output_fold >= 0.5).astype(int).flatten()
                        else:
                            raise ValueError("Unexpected raw prediction output shape for classification task during fold processing.")
                    else:  # Binary, 1D raw output
                        y_pred_class_fold = (y_pred_i_raw_output_fold >= 0.5).astype(int)
                    
                    y_pred_inverse_transformed_fold = dataset.inverse_transform(y_pred_class_fold)
                else:  # Regression
                    # For regression, raw output is the prediction; inverse_transform handles any scaling
                    y_pred_inverse_transformed_fold = dataset.inverse_transform(y_pred_i_raw_output_fold)
                
                all_preds_inverse_transformed_for_saving.append(y_pred_inverse_transformed_fold)
                # Evaluate scores for this fold using original y_true and original y_pred
                current_fold_scores = model_manager.evaluate(y_true_for_eval_and_save, y_pred_inverse_transformed_fold, metrics)
                fold_scores.append(current_fold_scores)

            # Compute mean prediction across all folds (from raw model outputs before class conversion)
            mean_raw_pred = np.mean(np.array(raw_preds_model_output_folds), axis=0)
            mean_pred_class = None
            if task == 'classification':
                if mean_raw_pred.ndim > 1:
                    if mean_raw_pred.shape[-1] == dataset.num_classes and dataset.num_classes > 1:
                        mean_pred_class = np.argmax(mean_raw_pred, axis=-1)
                    elif mean_raw_pred.shape[-1] == 1 or dataset.num_classes == 1:
                        mean_pred_class = (mean_raw_pred >= 0.5).astype(int).flatten()
                    else:
                        raise ValueError("Unexpected mean raw prediction output shape for classification task.")
                else:
                    mean_pred_class = (mean_raw_pred >= 0.5).astype(int)
                mean_pred_inverse_transformed = dataset.inverse_transform(mean_pred_class)
            else:  # Regression
                mean_pred_inverse_transformed = dataset.inverse_transform(mean_raw_pred)

            # Identify the best fold based on the first metric
            metric_to_use_for_best_fold = metrics[0]
            # Ensure scores for metric_to_use_for_best_fold are numeric, default if not found
            fold_metric_values = np.array([
                s.get(metric_to_use_for_best_fold, 0 if task == 'classification' else np.inf)
                for s in fold_scores if isinstance(s.get(metric_to_use_for_best_fold), (int, float))
            ])
            if len(fold_metric_values) == 0:  # Fallback if metric not found or not numeric
                fold_metric_values = np.array([0 if task == 'classification' else np.inf] * len(fold_scores))

            best_fold_index = np.argmax(fold_metric_values) if task == 'classification' else np.argmin(fold_metric_values)
            
            best_raw_pred_fold_output = raw_preds_model_output_folds[best_fold_index]
            best_pred_class_fold = None
            if task == 'classification':
                if best_raw_pred_fold_output.ndim > 1:
                    if best_raw_pred_fold_output.shape[-1] == dataset.num_classes and dataset.num_classes > 1:
                        best_pred_class_fold = np.argmax(best_raw_pred_fold_output, axis=-1)
                    elif best_raw_pred_fold_output.shape[-1] == 1 or dataset.num_classes == 1:
                        best_pred_class_fold = (best_raw_pred_fold_output >= 0.5).astype(int).flatten()
                    else:
                        raise ValueError("Unexpected best raw prediction output shape for classification task.")
                else:
                    best_pred_class_fold = (best_raw_pred_fold_output >= 0.5).astype(int)
                best_pred_inverse_transformed = dataset.inverse_transform(best_pred_class_fold)
            else:  # Regression
                best_pred_inverse_transformed = dataset.inverse_transform(best_raw_pred_fold_output)
            best_fold_overall_scores = fold_scores[best_fold_index]

            # Weighting logic for weighted average prediction
            weights_for_avg = None
            if task == 'classification':  # Higher score is better
                min_metric_val = np.min(fold_metric_values)
                adjusted_scores_for_weights = fold_metric_values - min_metric_val if min_metric_val < 0 else fold_metric_values
                total_score_for_weights = np.sum(adjusted_scores_for_weights)
                weights_for_avg = adjusted_scores_for_weights / total_score_for_weights if total_score_for_weights > 0 else np.ones_like(fold_metric_values) / len(fold_metric_values)
            else:  # Regression, lower score is better, so invert for weighting
                epsilon = 1e-8
                safe_scores_for_weights = np.maximum(fold_metric_values, epsilon)  # Avoid division by zero or issues with non-positive scores
                inverse_scores_for_weights = 1.0 / safe_scores_for_weights
                sum_inverse_scores = np.sum(inverse_scores_for_weights)
                weights_for_avg = inverse_scores_for_weights / sum_inverse_scores if sum_inverse_scores > 0 else np.ones_like(fold_metric_values) / len(fold_metric_values)
            
            weighted_avg_raw_pred = np.average(np.array(raw_preds_model_output_folds), axis=0, weights=weights_for_avg)
            weighted_avg_pred_class = None
            if task == 'classification':
                if weighted_avg_raw_pred.ndim > 1:
                    if weighted_avg_raw_pred.shape[-1] == dataset.num_classes and dataset.num_classes > 1:
                        weighted_avg_pred_class = np.argmax(weighted_avg_raw_pred, axis=-1)
                    elif weighted_avg_raw_pred.shape[-1] == 1 or dataset.num_classes == 1:
                        weighted_avg_pred_class = (weighted_avg_raw_pred >= 0.5).astype(int).flatten()
                    else:
                        raise ValueError("Unexpected weighted average raw prediction output shape for classification task.")
                else:
                    weighted_avg_pred_class = (weighted_avg_raw_pred >= 0.5).astype(int)
                weighted_avg_pred_inverse_transformed = dataset.inverse_transform(weighted_avg_pred_class)
            else:  # Regression
                weighted_avg_pred_inverse_transformed = dataset.inverse_transform(weighted_avg_raw_pred)

            # Evaluate mean, best, and weighted predictions
            mean_overall_scores = model_manager.evaluate(y_true_for_eval_and_save, mean_pred_inverse_transformed, metrics)
            # best_scores are already calculated as best_fold_overall_scores
            weighted_overall_scores = model_manager.evaluate(y_true_for_eval_and_save, weighted_avg_pred_inverse_transformed, metrics)
            
            all_preds_inverse_transformed_for_saving.extend([mean_pred_inverse_transformed, best_pred_inverse_transformed, weighted_avg_pred_inverse_transformed])
            # The fold_scores list already contains dicts of scores for each fold.
            # best_fold_overall_scores is the dict for the best fold.
            all_scores_to_save = fold_scores + [mean_overall_scores, best_fold_overall_scores, weighted_overall_scores]

            self.manager.save_results(model_manager, all_preds_inverse_transformed_for_saving, y_true_for_eval_and_save, metrics, best_params, all_scores_to_save)
            return all_preds_inverse_transformed_for_saving, all_scores_to_save, best_params

        else:  # Handling single prediction (no folds)
            y_pred_single_raw_output = y_pred_raw_outputs  # This is the direct model output
            
            y_pred_class_single = None
            if task == 'classification':
                if y_pred_single_raw_output.ndim > 1:
                    if y_pred_single_raw_output.shape[-1] == dataset.num_classes and dataset.num_classes > 1:  # Multi-class
                        y_pred_class_single = np.argmax(y_pred_single_raw_output, axis=-1)
                    elif y_pred_single_raw_output.shape[-1] == 1 or dataset.num_classes == 1:  # Binary
                        y_pred_class_single = (y_pred_single_raw_output >= 0.5).astype(int).flatten()
                    else:
                        raise ValueError("Unexpected raw prediction output shape for single classification task.")
                else:  # Binary, 1D raw output
                    y_pred_class_single = (y_pred_single_raw_output >= 0.5).astype(int)
                
                y_pred_inverse_transformed_single = dataset.inverse_transform(y_pred_class_single)
            else:  # Regression
                y_pred_inverse_transformed_single = dataset.inverse_transform(y_pred_single_raw_output)
            
            # Evaluate scores using original y_true and original y_pred
            final_scores = model_manager.evaluate(y_true_for_eval_and_save, y_pred_inverse_transformed_single, metrics)
            self.manager.save_results(model_manager, y_pred_inverse_transformed_single, y_true_for_eval_and_save, metrics, best_params, [final_scores])
            return y_pred_inverse_transformed_single, [final_scores], best_params

    def _train(self, model_manager, dataset, training_params, metrics, task):
        self.logger.info("Training the model")
        model_manager.train(dataset, training_params=training_params, metrics=metrics)
        model_manager.save_model(os.path.join(self.manager.experiment_path, "model"))
        self.logger.info("Saved model to %s", self.manager.experiment_path)
        return self._evaluate_and_save_results(model_manager, dataset, metrics, task=task)

    def _predict(self, model_manager, dataset, metrics, task):
        self.logger.info("Predicting using the model")
        return self._evaluate_and_save_results(model_manager, dataset, metrics, task=task)

    def _fine_tune(self, model_manager, dataset, finetune_params, training_params, metrics, task):
        self.logger.info("Finetuning the model")
        finetuner_type = finetune_params.get('tuner', 'optuna')
        finetuner = FineTunerFactory.get_fine_tuner(finetuner_type, model_manager)
        best_params = finetuner.finetune(dataset, finetune_params, training_params, metrics=metrics, task=task)
        model_manager.save_model(os.path.join(self.manager.experiment_path, "model"))
        return self._evaluate_and_save_results(model_manager, dataset, metrics, best_params, task=task)



from ._random_augmentation import (Random_X_Operation,
                                   Rotate_Translate)
from ._spline_augmentation import (Spline_Curve_Simplification,
                                   Spline_X_Simplification,
                                   Spline_Y_Perturbations,
                                   Spline_X_Perturbations,
                                   Spline_Smoothing)
from .augmenter import Augmenter, IdentityAugmenter

from sklearn.preprocessing import FunctionTransformer as IdentityTransformer
from sklearn.preprocessing import RobustScaler as RobustNormalVariate
from sklearn.preprocessing import StandardScaler as StandardNormalVariate

from ._nirs import (Haar, MultiplicativeScatterCorrection, SavitzkyGolay, Wavelet, msc, savgol, wavelet_transform)
from ._scaler import (Derivate, Normalize, SimpleScale, derivate, norml, spl_norml)
from ._standard import Baseline, Detrend, Gaussian, baseline, detrend, gaussian
from ._preparation import CropTransformer, ResampleTransformer
from .augmenter import Augmenter, IdentityAugmenter


__all__ = [
    "Spline_Smoothing",
    "Spline_X_Perturbations",
    "Spline_Y_Perturbations",
    "Spline_X_Simplification",
    "Spline_Curve_Simplification",
    "Rotate_Translate",
    "Random_X_Operation",
    "Augmenter",
    "IdentityAugmenter",
    "IdentityTransformer",  # sklearn.preprocessing.FunctionTransformer alias
    "Baseline",
    "StandardNormalVariate",  # sklearn.preprocessing.StandardScaler alias
    "RobustNormalVariate",  # sklearn.preprocessing.RobusScaler alias
    "SavitzkyGolay",
    "Haar",
    "Normalize",
    "Detrend",
    "MultiplicativeScatterCorrection",
    "Derivate",
    "Gaussian",
    "Wavelet",
    "SimpleScale",
    "baseline",
    "savgol",
    "norml",
    "detrend",
    "msc",
    "wavelet_transform",
    "derivate",
    "spl_norml",
    "gaussian",
    "CropTransformer",
    "ResampleTransformer",
    "Augmenter",
    "IdentityAugmenter",
]



from pathlib import Path

def _s_(path: str) -> str:
    if path is None:
        return None
    return Path(path).as_posix()

def browse_folder(folder_path, global_params=None):
    print(">> Browsing", folder_path)

    config = {
        "train_x": None, "train_x_filter": None, "train_x_params": None,
        "train_y": None, "train_y_filter": None, "train_y_params": None,
        "train_group": None, "train_group_filter": None, "train_group_params": None,
        "train_params": None,
        "test_x": None, "test_x_filter": None, "test_x_params": None,
        "test_y": None, "test_y_filter": None, "test_y_params": None,
        "test_group": None, "test_group_filter": None, "test_group_params": None,
        "test_params": None,
        "global_params": global_params
    }

    files_re = {
        "train_x": ["Xcal", "X_cal", "Cal_X", "calX", "train_X", "trainX", "X_train", "Xtrain"],
        "test_x": ["Xval", "X_val", "val_X", "valX", "Xtest", "X_test", "test_X", "testX"],
        "train_y": ["Ycal", "Y_cal", "Cal_Y", "calY", "train_Y", "trainY", "Y_train", "Ytrain"],
        "test_y": ["Ytest", "Y_test", "test_Y", "testY", "Yval", "Y_val", "val_Y", "valY"],
        "train_group": ["Gcal", "G_cal", "Cal_G", "calG", "train_G", "trainG", "G_train", "Gtrain"],
        "test_group": ["Gtest", "G_test", "test_G", "testG", "Gval", "G_val", "val_G", "valG"],
    }

    dataset_dir = Path(folder_path)
    for key, patterns in files_re.items():
        matched_files = []
        for pattern in patterns:
            pattern_lower = pattern.lower()
            for file in dataset_dir.glob("*"):
                if pattern_lower in file.name.lower():
                    matched_files.append(str(file))

        if len(matched_files) > 1:
            print(f"Multiple {key} files found for {folder_path}.")
            # logging.warning("Multiple %s files found for %s.", key, dataset_name)
            continue
        if len(matched_files) == 0:
            print(f"No {key} file found for {folder_path}.")
            # logging.warning("No %s file found for %s.", key, dataset_name)
            continue

        config[key] = _s_(matched_files[0])

    return config


# def parse_selector(filter):
#     if isinstance(filter, str):
#         return {"np_slice": filter}
#     elif isinstance(filter, dict):
#         return {"from": filter.get("from", "start"), "to": filter.get("to", "end")}
#     elif isinstance(filter, list):
#         return {"cols": filter}


# def parse_XY(XY):
#     # print("Parsing XY", XY)
#     if isinstance(XY, dict):
#         return {"x": XY["X"], "y": XY["Y"], "params": XY.get("params")}
#     elif isinstance(XY, list) or isinstance(XY, tuple):
#         return {"x": XY[0], "y": XY[1]}


# def parse_file(file):
#     # print("Parsing file", file)
#     if isinstance(file, str):
#         return {"path": file}
#     elif isinstance(file, dict):
#         return {"path": file["path"], "filter": parse_selector(file.get("filter")), "params": file.get("params")}
#     elif isinstance(file, list) or isinstance(file, tuple):
#         if len(file) == 2:
#             return {"path": file[0], "filter": parse_selector(file[1]["filter"])}
#         elif len(file) == 3:
#             return {"path": file[0], "filter": parse_selector(file[1]["filter"]), "params": file[2]}


# def parse_file_or_selector(file):
#     if isinstance(file, str):
#         if bool(re.match("^(\s*-?\d+\s*)?(:\s*-?\d+\\s*)?(:\s*-?\d+\s*)?$", file)):
#             return parse_selector(file)
#         else:
#             return parse_file(file)
#     elif isinstance(file, dict):
#         if 'path' in file:
#             return parse_file(file)
#         else:
#             return parse_selector(file)
#     elif isinstance(file, list) or isinstance(file, tuple):
#         if isinstance(file[0], str):
#             return parse_file(file)
#         else:
#             return parse_selector(file)


# def format_config(train_XY=None, test_XY=None, global_params=None, train_x=None, test_x=None, train_y=None, test_y=None, train_params=None, test_params=None):

#     if (train_x is not None and train_XY is not None) or (test_x is not None and test_XY is not None):
#         print("ERROR: both XY and X,Y are provided")

#     if (train_x is not None and train_y is None) or (test_x is not None and test_y is None):
#         print("ERROR: both X and Y must be provided")

#     train_x_path, train_x_filter, train_x_params = None, None, None
#     train_y_path, train_y_filter, train_y_params = None, None, None
#     test_x_path, test_x_filter, test_x_params = None, None, None
#     test_y_path, test_y_filter, test_y_params = None, None, None
#     train_params, test_params = None, None

#     if train_XY is not None:
#         p_XY = parse_XY(train_XY)
#         train_params = p_XY.get("params")
#         file_x = parse_file(p_XY["x"])
#         train_x_path, train_x_filter, train_x_params = file_x["path"], file_x.get("filter"), file_x.get("params")
#         p_y = parse_file_or_selector(p_XY["y"])
#         if 'path' in p_y:
#             train_y_path, train_y_filter, train_y_params = p_y["path"], p_y.get("filter"), p_y.get("params")
#         else:
#             train_y_filter = p_y
#             train_y_path = "From X"

#     if test_XY is not None:
#         p_XY = parse_XY(test_XY)
#         test_params = p_XY.get("params")
#         file_x = parse_file(p_XY["x"])
#         test_x_path, test_x_filter, test_x_params = file_x["path"], file_x.get("filter"), file_x.get("params")
#         p_y = parse_file_or_selector(p_XY["y"])
#         if 'path' in p_y:
#             test_y_path, test_y_filter, test_y_params = p_y["path"], p_y.get("filter"), p_y.get("params")
#         else:
#             test_y_filter = p_y
#             test_y_path = "From X"

#     if train_x is not None:
#         p_x = parse_file(train_x)
#         train_x_path, train_x_filter, train_x_params = p_x["path"], p_x.get("filter"), p_x.get("params")

#         p_y = parse_file_or_selector(train_y)
#         if 'path' in p_y:
#             train_y_path, train_y_filter, train_y_params = p_y["path"], p_y.get("filter"), p_y.get("params")
#         else:
#             train_y_filter = p_y
#             train_y_path = "From X"

#     if test_x is not None:
#         p_x = parse_file(test_x)
#         test_x_path, test_x_filter, test_x_params = p_x["path"], p_x.get("filter"), p_x.get("params")

#         p_y = parse_file_or_selector(test_y)
#         if 'path' in p_y:
#             test_y_path, test_y_filter, test_y_params = p_y["path"], p_y.get("filter"), p_y.get("params")
#         else:
#             test_y_filter = p_y
#             test_y_path = "From X"

#     return {
#         "train_x": _s_(train_x_path), "train_x_filter": train_x_filter, "train_x_params": train_x_params,
#         "train_y": _s_(train_y_path), "train_y_filter": train_y_filter, "train_y_params": train_y_params,
#         "train_params": train_params,
#         "test_x": _s_(test_x_path), "test_x_filter": test_x_filter, "test_x_params": test_x_params,
#         "test_y": _s_(test_y_path), "test_y_filter": test_y_filter, "test_y_params": test_y_params,
#         "test_params": test_params,
#         "global_params": global_params
#     }


def parse_config(data_config):
    if isinstance(data_config, str):
        return browse_folder(data_config)
    elif isinstance(data_config, dict):
        if "path" in data_config:  # If path is present, browse folder
            return browse_folder(data_config["path"], data_config.get("params"))
        else:  # Otherwise, assume it's an already parsed config dictionary
            # TODO: Add more robust validation here if needed in the future
            # For now, if it's a dict without 'path', assume it's usable by get_dataset
            required_keys_pattern = ['train_x', 'test_x']  # Basic check
            if all(key in data_config for key in required_keys_pattern):
                return data_config
            else:
                print(f"CONFIG ERROR _ obj >> Missing one of {required_keys_pattern} in dict: {data_config}")
                return None
    # If data_config is not a string or a recognized dictionary structure, return None
    print(f"CONFIG ERROR _ unsupported type or structure >> {type(data_config)}: {data_config}")
    return None
